r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Lookups
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from typing import Any, Dict, Optional, Union
from twilio.base import values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version


class BucketInstance(InstanceResource):

    class RateLimitRequest(object):
        """
        :ivar limit: Limit of requests for the bucket
        :ivar ttl: Time to live of the rule
        """

        def __init__(self, payload: Dict[str, Any]):

            self.limit: Optional[int] = payload.get("limit")
            self.ttl: Optional[int] = payload.get("ttl")

        def to_dict(self):
            return {
                "limit": self.limit,
                "ttl": self.ttl,
            }

    """
    :ivar code: Twilio-specific error code
    :ivar message: Error message
    :ivar more_info: Link to Error Code References
    :ivar status: HTTP response status code
    :ivar field: Limit of requests for the bucket
    :ivar limit: Limit of requests for the bucket
    :ivar bucket: Name of the bucket
    :ivar owner: Owner of the rule
    :ivar ttl: Time to live of the rule
    """

    def __init__(
        self,
        version: Version,
        payload: Dict[str, Any],
        field: Optional[str] = None,
        bucket: Optional[str] = None,
    ):
        super().__init__(version)

        self.code: Optional[int] = payload.get("code")
        self.message: Optional[str] = payload.get("message")
        self.more_info: Optional[str] = payload.get("more_info")
        self.status: Optional[int] = payload.get("status")
        self.field: Optional[str] = payload.get("field")
        self.limit: Optional[int] = payload.get("limit")
        self.bucket: Optional[str] = payload.get("bucket")
        self.owner: Optional[str] = payload.get("owner")
        self.ttl: Optional[int] = payload.get("ttl")

        self._solution = {
            "field": field or self.field,
            "bucket": bucket or self.bucket,
        }
        self._context: Optional[BucketContext] = None

    @property
    def _proxy(self) -> "BucketContext":
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: BucketContext for this BucketInstance
        """
        if self._context is None:
            self._context = BucketContext(
                self._version,
                field=self._solution["field"],
                bucket=self._solution["bucket"],
            )
        return self._context

    def delete(self) -> bool:
        """
        Deletes the BucketInstance


        :returns: True if delete succeeds, False otherwise
        """
        return self._proxy.delete()

    async def delete_async(self) -> bool:
        """
        Asynchronous coroutine that deletes the BucketInstance


        :returns: True if delete succeeds, False otherwise
        """
        return await self._proxy.delete_async()

    def fetch(self) -> "BucketInstance":
        """
        Fetch the BucketInstance


        :returns: The fetched BucketInstance
        """
        return self._proxy.fetch()

    async def fetch_async(self) -> "BucketInstance":
        """
        Asynchronous coroutine to fetch the BucketInstance


        :returns: The fetched BucketInstance
        """
        return await self._proxy.fetch_async()

    def update(
        self, rate_limit_request: Union[RateLimitRequest, object] = values.unset
    ) -> "BucketInstance":
        """
        Update the BucketInstance

        :param rate_limit_request:

        :returns: The updated BucketInstance
        """
        return self._proxy.update(
            rate_limit_request=rate_limit_request,
        )

    async def update_async(
        self, rate_limit_request: Union[RateLimitRequest, object] = values.unset
    ) -> "BucketInstance":
        """
        Asynchronous coroutine to update the BucketInstance

        :param rate_limit_request:

        :returns: The updated BucketInstance
        """
        return await self._proxy.update_async(
            rate_limit_request=rate_limit_request,
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Lookups.V2.BucketInstance {}>".format(context)


class BucketContext(InstanceContext):

    class RateLimitRequest(object):
        """
        :ivar limit: Limit of requests for the bucket
        :ivar ttl: Time to live of the rule
        """

        def __init__(self, payload: Dict[str, Any]):

            self.limit: Optional[int] = payload.get("limit")
            self.ttl: Optional[int] = payload.get("ttl")

        def to_dict(self):
            return {
                "limit": self.limit,
                "ttl": self.ttl,
            }

    def __init__(self, version: Version, field: str, bucket: str):
        """
        Initialize the BucketContext

        :param version: Version that contains the resource
        :param field: field name
        :param bucket: bucket name
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "field": field,
            "bucket": bucket,
        }
        self._uri = "/RateLimits/Fields/{field}/Bucket/{bucket}".format(
            **self._solution
        )

    def delete(self) -> bool:
        """
        Deletes the BucketInstance


        :returns: True if delete succeeds, False otherwise
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        return self._version.delete(method="DELETE", uri=self._uri, headers=headers)

    async def delete_async(self) -> bool:
        """
        Asynchronous coroutine that deletes the BucketInstance


        :returns: True if delete succeeds, False otherwise
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        return await self._version.delete_async(
            method="DELETE", uri=self._uri, headers=headers
        )

    def fetch(self) -> BucketInstance:
        """
        Fetch the BucketInstance


        :returns: The fetched BucketInstance
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = self._version.fetch(method="GET", uri=self._uri, headers=headers)

        return BucketInstance(
            self._version,
            payload,
            field=self._solution["field"],
            bucket=self._solution["bucket"],
        )

    async def fetch_async(self) -> BucketInstance:
        """
        Asynchronous coroutine to fetch the BucketInstance


        :returns: The fetched BucketInstance
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = await self._version.fetch_async(
            method="GET", uri=self._uri, headers=headers
        )

        return BucketInstance(
            self._version,
            payload,
            field=self._solution["field"],
            bucket=self._solution["bucket"],
        )

    def update(
        self, rate_limit_request: Union[RateLimitRequest, object] = values.unset
    ) -> BucketInstance:
        """
        Update the BucketInstance

        :param rate_limit_request:

        :returns: The updated BucketInstance
        """
        data = rate_limit_request.to_dict()

        headers = values.of({})

        headers["Content-Type"] = "application/json"

        headers["Accept"] = "application/json"

        payload = self._version.update(
            method="PUT", uri=self._uri, data=data, headers=headers
        )

        return BucketInstance(
            self._version,
            payload,
            field=self._solution["field"],
            bucket=self._solution["bucket"],
        )

    async def update_async(
        self, rate_limit_request: Union[RateLimitRequest, object] = values.unset
    ) -> BucketInstance:
        """
        Asynchronous coroutine to update the BucketInstance

        :param rate_limit_request:

        :returns: The updated BucketInstance
        """
        data = rate_limit_request.to_dict()

        headers = values.of({})

        headers["Content-Type"] = "application/json"

        headers["Accept"] = "application/json"

        payload = await self._version.update_async(
            method="PUT", uri=self._uri, data=data, headers=headers
        )

        return BucketInstance(
            self._version,
            payload,
            field=self._solution["field"],
            bucket=self._solution["bucket"],
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Lookups.V2.BucketContext {}>".format(context)


class BucketList(ListResource):

    class RateLimitRequest(object):
        """
        :ivar limit: Limit of requests for the bucket
        :ivar ttl: Time to live of the rule
        """

        def __init__(self, payload: Dict[str, Any]):

            self.limit: Optional[int] = payload.get("limit")
            self.ttl: Optional[int] = payload.get("ttl")

        def to_dict(self):
            return {
                "limit": self.limit,
                "ttl": self.ttl,
            }

    def __init__(self, version: Version):
        """
        Initialize the BucketList

        :param version: Version that contains the resource

        """
        super().__init__(version)

    def get(self, field: str, bucket: str) -> BucketContext:
        """
        Constructs a BucketContext

        :param field: field name
        :param bucket: bucket name
        """
        return BucketContext(self._version, field=field, bucket=bucket)

    def __call__(self, field: str, bucket: str) -> BucketContext:
        """
        Constructs a BucketContext

        :param field: field name
        :param bucket: bucket name
        """
        return BucketContext(self._version, field=field, bucket=bucket)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Lookups.V2.BucketList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Numbers
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from typing import Optional
from twilio.base.version import Version
from twilio.base.domain import Domain
from twilio.rest.numbers.v3.hosted_number_order import HostedNumberOrderList


class V3(Version):

    def __init__(self, domain: Domain):
        """
        Initialize the V3 version of Numbers

        :param domain: The Twilio.numbers domain
        """
        super().__init__(domain, "v3")
        self._hosted_number_orders: Optional[HostedNumberOrderList] = None

    @property
    def hosted_number_orders(self) -> HostedNumberOrderList:
        if self._hosted_number_orders is None:
            self._hosted_number_orders = HostedNumberOrderList(self)
        return self._hosted_number_orders

    def __repr__(self) -> str:
        """
        Provide a friendly representation
        :returns: Machine friendly representation
        """
        return "<Twilio.Numbers.V3>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Events
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from datetime import datetime
from typing import Any, Dict, Optional
from twilio.base import deserialize, values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version

from twilio.rest.events.v1.schema.schema_version import SchemaVersionList


class SchemaInstance(InstanceResource):
    """
    :ivar id: The unique identifier of the schema. Each schema can have multiple versions, that share the same id.
    :ivar url: The URL of this resource.
    :ivar links: Contains a dictionary of URL links to nested resources of this schema.
    :ivar latest_version_date_created: The date that the latest schema version was created, given in ISO 8601 format.
    :ivar latest_version: The latest version published of this schema.
    """

    def __init__(
        self, version: Version, payload: Dict[str, Any], id: Optional[str] = None
    ):
        super().__init__(version)

        self.id: Optional[str] = payload.get("id")
        self.url: Optional[str] = payload.get("url")
        self.links: Optional[Dict[str, object]] = payload.get("links")
        self.latest_version_date_created: Optional[datetime] = (
            deserialize.iso8601_datetime(payload.get("latest_version_date_created"))
        )
        self.latest_version: Optional[int] = deserialize.integer(
            payload.get("latest_version")
        )

        self._solution = {
            "id": id or self.id,
        }
        self._context: Optional[SchemaContext] = None

    @property
    def _proxy(self) -> "SchemaContext":
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: SchemaContext for this SchemaInstance
        """
        if self._context is None:
            self._context = SchemaContext(
                self._version,
                id=self._solution["id"],
            )
        return self._context

    def fetch(self) -> "SchemaInstance":
        """
        Fetch the SchemaInstance


        :returns: The fetched SchemaInstance
        """
        return self._proxy.fetch()

    async def fetch_async(self) -> "SchemaInstance":
        """
        Asynchronous coroutine to fetch the SchemaInstance


        :returns: The fetched SchemaInstance
        """
        return await self._proxy.fetch_async()

    @property
    def versions(self) -> SchemaVersionList:
        """
        Access the versions
        """
        return self._proxy.versions

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Events.V1.SchemaInstance {}>".format(context)


class SchemaContext(InstanceContext):

    def __init__(self, version: Version, id: str):
        """
        Initialize the SchemaContext

        :param version: Version that contains the resource
        :param id: The unique identifier of the schema. Each schema can have multiple versions, that share the same id.
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "id": id,
        }
        self._uri = "/Schemas/{id}".format(**self._solution)

        self._versions: Optional[SchemaVersionList] = None

    def fetch(self) -> SchemaInstance:
        """
        Fetch the SchemaInstance


        :returns: The fetched SchemaInstance
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = self._version.fetch(method="GET", uri=self._uri, headers=headers)

        return SchemaInstance(
            self._version,
            payload,
            id=self._solution["id"],
        )

    async def fetch_async(self) -> SchemaInstance:
        """
        Asynchronous coroutine to fetch the SchemaInstance


        :returns: The fetched SchemaInstance
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = await self._version.fetch_async(
            method="GET", uri=self._uri, headers=headers
        )

        return SchemaInstance(
            self._version,
            payload,
            id=self._solution["id"],
        )

    @property
    def versions(self) -> SchemaVersionList:
        """
        Access the versions
        """
        if self._versions is None:
            self._versions = SchemaVersionList(
                self._version,
                self._solution["id"],
            )
        return self._versions

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Events.V1.SchemaContext {}>".format(context)


class SchemaList(ListResource):

    def __init__(self, version: Version):
        """
        Initialize the SchemaList

        :param version: Version that contains the resource

        """
        super().__init__(version)

    def get(self, id: str) -> SchemaContext:
        """
        Constructs a SchemaContext

        :param id: The unique identifier of the schema. Each schema can have multiple versions, that share the same id.
        """
        return SchemaContext(self._version, id=id)

    def __call__(self, id: str) -> SchemaContext:
        """
        Constructs a SchemaContext

        :param id: The unique identifier of the schema. Each schema can have multiple versions, that share the same id.
        """
        return SchemaContext(self._version, id=id)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Events.V1.SchemaList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Lookups
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from typing import Any, Dict, List, Optional, Union
from twilio.base import values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version


class PhoneNumberInstance(InstanceResource):

    class ValidationError(object):
        TOO_SHORT = "TOO_SHORT"
        TOO_LONG = "TOO_LONG"
        INVALID_BUT_POSSIBLE = "INVALID_BUT_POSSIBLE"
        INVALID_COUNTRY_CODE = "INVALID_COUNTRY_CODE"
        INVALID_LENGTH = "INVALID_LENGTH"
        NOT_A_NUMBER = "NOT_A_NUMBER"

    """
    :ivar calling_country_code: International dialing prefix of the phone number defined in the E.164 standard.
    :ivar country_code: The phone number's [ISO country code](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2).
    :ivar phone_number: The phone number in [E.164](https://www.twilio.com/docs/glossary/what-e164) format, which consists of a + followed by the country code and subscriber number.
    :ivar national_format: The phone number in [national format](https://en.wikipedia.org/wiki/National_conventions_for_writing_telephone_numbers).
    :ivar valid: Boolean which indicates if the phone number is in a valid range that can be freely assigned by a carrier to a user.
    :ivar validation_errors: Contains reasons why a phone number is invalid. Possible values: TOO_SHORT, TOO_LONG, INVALID_BUT_POSSIBLE, INVALID_COUNTRY_CODE, INVALID_LENGTH, NOT_A_NUMBER.
    :ivar caller_name: 
    :ivar sim_swap: 
    :ivar call_forwarding: 
    :ivar line_type_intelligence: 
    :ivar line_status: 
    :ivar identity_match: 
    :ivar reassigned_number: 
    :ivar sms_pumping_risk: 
    :ivar phone_number_quality_score: An object that contains information of a mobile phone number quality score. Quality score will return a risk score about the phone number.
    :ivar pre_fill: An object that contains pre fill information. pre_fill will return PII information associated with the phone number like first name, last name, address line, country code, state and postal code. 
    :ivar url: The absolute URL of the resource.
    """

    def __init__(
        self,
        version: Version,
        payload: Dict[str, Any],
        phone_number: Optional[str] = None,
    ):
        super().__init__(version)

        self.calling_country_code: Optional[str] = payload.get("calling_country_code")
        self.country_code: Optional[str] = payload.get("country_code")
        self.phone_number: Optional[str] = payload.get("phone_number")
        self.national_format: Optional[str] = payload.get("national_format")
        self.valid: Optional[bool] = payload.get("valid")
        self.validation_errors: Optional[List[str]] = payload.get("validation_errors")
        self.caller_name: Optional[str] = payload.get("caller_name")
        self.sim_swap: Optional[str] = payload.get("sim_swap")
        self.call_forwarding: Optional[str] = payload.get("call_forwarding")
        self.line_type_intelligence: Optional[str] = payload.get(
            "line_type_intelligence"
        )
        self.line_status: Optional[str] = payload.get("line_status")
        self.identity_match: Optional[str] = payload.get("identity_match")
        self.reassigned_number: Optional[str] = payload.get("reassigned_number")
        self.sms_pumping_risk: Optional[str] = payload.get("sms_pumping_risk")
        self.phone_number_quality_score: Optional[Dict[str, object]] = payload.get(
            "phone_number_quality_score"
        )
        self.pre_fill: Optional[Dict[str, object]] = payload.get("pre_fill")
        self.url: Optional[str] = payload.get("url")

        self._solution = {
            "phone_number": phone_number or self.phone_number,
        }
        self._context: Optional[PhoneNumberContext] = None

    @property
    def _proxy(self) -> "PhoneNumberContext":
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: PhoneNumberContext for this PhoneNumberInstance
        """
        if self._context is None:
            self._context = PhoneNumberContext(
                self._version,
                phone_number=self._solution["phone_number"],
            )
        return self._context

    def fetch(
        self,
        fields: Union[str, object] = values.unset,
        country_code: Union[str, object] = values.unset,
        first_name: Union[str, object] = values.unset,
        last_name: Union[str, object] = values.unset,
        address_line1: Union[str, object] = values.unset,
        address_line2: Union[str, object] = values.unset,
        city: Union[str, object] = values.unset,
        state: Union[str, object] = values.unset,
        postal_code: Union[str, object] = values.unset,
        address_country_code: Union[str, object] = values.unset,
        national_id: Union[str, object] = values.unset,
        date_of_birth: Union[str, object] = values.unset,
        last_verified_date: Union[str, object] = values.unset,
        verification_sid: Union[str, object] = values.unset,
        partner_sub_id: Union[str, object] = values.unset,
    ) -> "PhoneNumberInstance":
        """
        Fetch the PhoneNumberInstance

        :param fields: A comma-separated list of fields to return. Possible values are validation, caller_name, sim_swap, call_forwarding, line_status, line_type_intelligence, identity_match, reassigned_number, sms_pumping_risk, phone_number_quality_score, pre_fill.
        :param country_code: The [country code](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2) used if the phone number provided is in national format.
        :param first_name: User’s first name. This query parameter is only used (optionally) for identity_match package requests.
        :param last_name: User’s last name. This query parameter is only used (optionally) for identity_match package requests.
        :param address_line1: User’s first address line. This query parameter is only used (optionally) for identity_match package requests.
        :param address_line2: User’s second address line. This query parameter is only used (optionally) for identity_match package requests.
        :param city: User’s city. This query parameter is only used (optionally) for identity_match package requests.
        :param state: User’s country subdivision, such as state, province, or locality. This query parameter is only used (optionally) for identity_match package requests.
        :param postal_code: User’s postal zip code. This query parameter is only used (optionally) for identity_match package requests.
        :param address_country_code: User’s country, up to two characters. This query parameter is only used (optionally) for identity_match package requests.
        :param national_id: User’s national ID, such as SSN or Passport ID. This query parameter is only used (optionally) for identity_match package requests.
        :param date_of_birth: User’s date of birth, in YYYYMMDD format. This query parameter is only used (optionally) for identity_match package requests.
        :param last_verified_date: The date you obtained consent to call or text the end-user of the phone number or a date on which you are reasonably certain that the end-user could still be reached at that number. This query parameter is only used (optionally) for reassigned_number package requests.
        :param verification_sid: The unique identifier associated with a verification process through verify API. This query parameter is only used (optionally) for pre_fill package requests.
        :param partner_sub_id: The optional partnerSubId parameter to provide context for your sub-accounts, tenantIDs, sender IDs or other segmentation, enhancing the accuracy of the risk analysis.

        :returns: The fetched PhoneNumberInstance
        """
        return self._proxy.fetch(
            fields=fields,
            country_code=country_code,
            first_name=first_name,
            last_name=last_name,
            address_line1=address_line1,
            address_line2=address_line2,
            city=city,
            state=state,
            postal_code=postal_code,
            address_country_code=address_country_code,
            national_id=national_id,
            date_of_birth=date_of_birth,
            last_verified_date=last_verified_date,
            verification_sid=verification_sid,
            partner_sub_id=partner_sub_id,
        )

    async def fetch_async(
        self,
        fields: Union[str, object] = values.unset,
        country_code: Union[str, object] = values.unset,
        first_name: Union[str, object] = values.unset,
        last_name: Union[str, object] = values.unset,
        address_line1: Union[str, object] = values.unset,
        address_line2: Union[str, object] = values.unset,
        city: Union[str, object] = values.unset,
        state: Union[str, object] = values.unset,
        postal_code: Union[str, object] = values.unset,
        address_country_code: Union[str, object] = values.unset,
        national_id: Union[str, object] = values.unset,
        date_of_birth: Union[str, object] = values.unset,
        last_verified_date: Union[str, object] = values.unset,
        verification_sid: Union[str, object] = values.unset,
        partner_sub_id: Union[str, object] = values.unset,
    ) -> "PhoneNumberInstance":
        """
        Asynchronous coroutine to fetch the PhoneNumberInstance

        :param fields: A comma-separated list of fields to return. Possible values are validation, caller_name, sim_swap, call_forwarding, line_status, line_type_intelligence, identity_match, reassigned_number, sms_pumping_risk, phone_number_quality_score, pre_fill.
        :param country_code: The [country code](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2) used if the phone number provided is in national format.
        :param first_name: User’s first name. This query parameter is only used (optionally) for identity_match package requests.
        :param last_name: User’s last name. This query parameter is only used (optionally) for identity_match package requests.
        :param address_line1: User’s first address line. This query parameter is only used (optionally) for identity_match package requests.
        :param address_line2: User’s second address line. This query parameter is only used (optionally) for identity_match package requests.
        :param city: User’s city. This query parameter is only used (optionally) for identity_match package requests.
        :param state: User’s country subdivision, such as state, province, or locality. This query parameter is only used (optionally) for identity_match package requests.
        :param postal_code: User’s postal zip code. This query parameter is only used (optionally) for identity_match package requests.
        :param address_country_code: User’s country, up to two characters. This query parameter is only used (optionally) for identity_match package requests.
        :param national_id: User’s national ID, such as SSN or Passport ID. This query parameter is only used (optionally) for identity_match package requests.
        :param date_of_birth: User’s date of birth, in YYYYMMDD format. This query parameter is only used (optionally) for identity_match package requests.
        :param last_verified_date: The date you obtained consent to call or text the end-user of the phone number or a date on which you are reasonably certain that the end-user could still be reached at that number. This query parameter is only used (optionally) for reassigned_number package requests.
        :param verification_sid: The unique identifier associated with a verification process through verify API. This query parameter is only used (optionally) for pre_fill package requests.
        :param partner_sub_id: The optional partnerSubId parameter to provide context for your sub-accounts, tenantIDs, sender IDs or other segmentation, enhancing the accuracy of the risk analysis.

        :returns: The fetched PhoneNumberInstance
        """
        return await self._proxy.fetch_async(
            fields=fields,
            country_code=country_code,
            first_name=first_name,
            last_name=last_name,
            address_line1=address_line1,
            address_line2=address_line2,
            city=city,
            state=state,
            postal_code=postal_code,
            address_country_code=address_country_code,
            national_id=national_id,
            date_of_birth=date_of_birth,
            last_verified_date=last_verified_date,
            verification_sid=verification_sid,
            partner_sub_id=partner_sub_id,
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Lookups.V2.PhoneNumberInstance {}>".format(context)


class PhoneNumberContext(InstanceContext):

    def __init__(self, version: Version, phone_number: str):
        """
        Initialize the PhoneNumberContext

        :param version: Version that contains the resource
        :param phone_number: The phone number to lookup in E.164 or national format. Default country code is +1 (North America).
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "phone_number": phone_number,
        }
        self._uri = "/PhoneNumbers/{phone_number}".format(**self._solution)

    def fetch(
        self,
        fields: Union[str, object] = values.unset,
        country_code: Union[str, object] = values.unset,
        first_name: Union[str, object] = values.unset,
        last_name: Union[str, object] = values.unset,
        address_line1: Union[str, object] = values.unset,
        address_line2: Union[str, object] = values.unset,
        city: Union[str, object] = values.unset,
        state: Union[str, object] = values.unset,
        postal_code: Union[str, object] = values.unset,
        address_country_code: Union[str, object] = values.unset,
        national_id: Union[str, object] = values.unset,
        date_of_birth: Union[str, object] = values.unset,
        last_verified_date: Union[str, object] = values.unset,
        verification_sid: Union[str, object] = values.unset,
        partner_sub_id: Union[str, object] = values.unset,
    ) -> PhoneNumberInstance:
        """
        Fetch the PhoneNumberInstance

        :param fields: A comma-separated list of fields to return. Possible values are validation, caller_name, sim_swap, call_forwarding, line_status, line_type_intelligence, identity_match, reassigned_number, sms_pumping_risk, phone_number_quality_score, pre_fill.
        :param country_code: The [country code](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2) used if the phone number provided is in national format.
        :param first_name: User’s first name. This query parameter is only used (optionally) for identity_match package requests.
        :param last_name: User’s last name. This query parameter is only used (optionally) for identity_match package requests.
        :param address_line1: User’s first address line. This query parameter is only used (optionally) for identity_match package requests.
        :param address_line2: User’s second address line. This query parameter is only used (optionally) for identity_match package requests.
        :param city: User’s city. This query parameter is only used (optionally) for identity_match package requests.
        :param state: User’s country subdivision, such as state, province, or locality. This query parameter is only used (optionally) for identity_match package requests.
        :param postal_code: User’s postal zip code. This query parameter is only used (optionally) for identity_match package requests.
        :param address_country_code: User’s country, up to two characters. This query parameter is only used (optionally) for identity_match package requests.
        :param national_id: User’s national ID, such as SSN or Passport ID. This query parameter is only used (optionally) for identity_match package requests.
        :param date_of_birth: User’s date of birth, in YYYYMMDD format. This query parameter is only used (optionally) for identity_match package requests.
        :param last_verified_date: The date you obtained consent to call or text the end-user of the phone number or a date on which you are reasonably certain that the end-user could still be reached at that number. This query parameter is only used (optionally) for reassigned_number package requests.
        :param verification_sid: The unique identifier associated with a verification process through verify API. This query parameter is only used (optionally) for pre_fill package requests.
        :param partner_sub_id: The optional partnerSubId parameter to provide context for your sub-accounts, tenantIDs, sender IDs or other segmentation, enhancing the accuracy of the risk analysis.

        :returns: The fetched PhoneNumberInstance
        """

        params = values.of(
            {
                "Fields": fields,
                "CountryCode": country_code,
                "FirstName": first_name,
                "LastName": last_name,
                "AddressLine1": address_line1,
                "AddressLine2": address_line2,
                "City": city,
                "State": state,
                "PostalCode": postal_code,
                "AddressCountryCode": address_country_code,
                "NationalId": national_id,
                "DateOfBirth": date_of_birth,
                "LastVerifiedDate": last_verified_date,
                "VerificationSid": verification_sid,
                "PartnerSubId": partner_sub_id,
            }
        )

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = self._version.fetch(
            method="GET", uri=self._uri, params=params, headers=headers
        )

        return PhoneNumberInstance(
            self._version,
            payload,
            phone_number=self._solution["phone_number"],
        )

    async def fetch_async(
        self,
        fields: Union[str, object] = values.unset,
        country_code: Union[str, object] = values.unset,
        first_name: Union[str, object] = values.unset,
        last_name: Union[str, object] = values.unset,
        address_line1: Union[str, object] = values.unset,
        address_line2: Union[str, object] = values.unset,
        city: Union[str, object] = values.unset,
        state: Union[str, object] = values.unset,
        postal_code: Union[str, object] = values.unset,
        address_country_code: Union[str, object] = values.unset,
        national_id: Union[str, object] = values.unset,
        date_of_birth: Union[str, object] = values.unset,
        last_verified_date: Union[str, object] = values.unset,
        verification_sid: Union[str, object] = values.unset,
        partner_sub_id: Union[str, object] = values.unset,
    ) -> PhoneNumberInstance:
        """
        Asynchronous coroutine to fetch the PhoneNumberInstance

        :param fields: A comma-separated list of fields to return. Possible values are validation, caller_name, sim_swap, call_forwarding, line_status, line_type_intelligence, identity_match, reassigned_number, sms_pumping_risk, phone_number_quality_score, pre_fill.
        :param country_code: The [country code](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2) used if the phone number provided is in national format.
        :param first_name: User’s first name. This query parameter is only used (optionally) for identity_match package requests.
        :param last_name: User’s last name. This query parameter is only used (optionally) for identity_match package requests.
        :param address_line1: User’s first address line. This query parameter is only used (optionally) for identity_match package requests.
        :param address_line2: User’s second address line. This query parameter is only used (optionally) for identity_match package requests.
        :param city: User’s city. This query parameter is only used (optionally) for identity_match package requests.
        :param state: User’s country subdivision, such as state, province, or locality. This query parameter is only used (optionally) for identity_match package requests.
        :param postal_code: User’s postal zip code. This query parameter is only used (optionally) for identity_match package requests.
        :param address_country_code: User’s country, up to two characters. This query parameter is only used (optionally) for identity_match package requests.
        :param national_id: User’s national ID, such as SSN or Passport ID. This query parameter is only used (optionally) for identity_match package requests.
        :param date_of_birth: User’s date of birth, in YYYYMMDD format. This query parameter is only used (optionally) for identity_match package requests.
        :param last_verified_date: The date you obtained consent to call or text the end-user of the phone number or a date on which you are reasonably certain that the end-user could still be reached at that number. This query parameter is only used (optionally) for reassigned_number package requests.
        :param verification_sid: The unique identifier associated with a verification process through verify API. This query parameter is only used (optionally) for pre_fill package requests.
        :param partner_sub_id: The optional partnerSubId parameter to provide context for your sub-accounts, tenantIDs, sender IDs or other segmentation, enhancing the accuracy of the risk analysis.

        :returns: The fetched PhoneNumberInstance
        """

        params = values.of(
            {
                "Fields": fields,
                "CountryCode": country_code,
                "FirstName": first_name,
                "LastName": last_name,
                "AddressLine1": address_line1,
                "AddressLine2": address_line2,
                "City": city,
                "State": state,
                "PostalCode": postal_code,
                "AddressCountryCode": address_country_code,
                "NationalId": national_id,
                "DateOfBirth": date_of_birth,
                "LastVerifiedDate": last_verified_date,
                "VerificationSid": verification_sid,
                "PartnerSubId": partner_sub_id,
            }
        )

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = await self._version.fetch_async(
            method="GET", uri=self._uri, params=params, headers=headers
        )

        return PhoneNumberInstance(
            self._version,
            payload,
            phone_number=self._solution["phone_number"],
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Lookups.V2.PhoneNumberContext {}>".format(context)


class PhoneNumberList(ListResource):

    def __init__(self, version: Version):
        """
        Initialize the PhoneNumberList

        :param version: Version that contains the resource

        """
        super().__init__(version)

    def get(self, phone_number: str) -> PhoneNumberContext:
        """
        Constructs a PhoneNumberContext

        :param phone_number: The phone number to lookup in E.164 or national format. Default country code is +1 (North America).
        """
        return PhoneNumberContext(self._version, phone_number=phone_number)

    def __call__(self, phone_number: str) -> PhoneNumberContext:
        """
        Constructs a PhoneNumberContext

        :param phone_number: The phone number to lookup in E.164 or national format. Default country code is +1 (North America).
        """
        return PhoneNumberContext(self._version, phone_number=phone_number)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Lookups.V2.PhoneNumberList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Lookups
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from typing import Optional
from twilio.base.version import Version
from twilio.base.domain import Domain
from twilio.rest.lookups.v2.bucket import BucketList
from twilio.rest.lookups.v2.lookup_override import LookupOverrideList
from twilio.rest.lookups.v2.phone_number import PhoneNumberList
from twilio.rest.lookups.v2.query import QueryList
from twilio.rest.lookups.v2.rate_limit import RateLimitList


class V2(Version):

    def __init__(self, domain: Domain):
        """
        Initialize the V2 version of Lookups

        :param domain: The Twilio.lookups domain
        """
        super().__init__(domain, "v2")
        self._bucket: Optional[BucketList] = None
        self._lookup_overrides: Optional[LookupOverrideList] = None
        self._phone_numbers: Optional[PhoneNumberList] = None
        self._query: Optional[QueryList] = None
        self._rate_limits: Optional[RateLimitList] = None

    @property
    def bucket(self) -> BucketList:
        if self._bucket is None:
            self._bucket = BucketList(self)
        return self._bucket

    @property
    def lookup_overrides(self) -> LookupOverrideList:
        if self._lookup_overrides is None:
            self._lookup_overrides = LookupOverrideList(self)
        return self._lookup_overrides

    @property
    def phone_numbers(self) -> PhoneNumberList:
        if self._phone_numbers is None:
            self._phone_numbers = PhoneNumberList(self)
        return self._phone_numbers

    @property
    def query(self) -> QueryList:
        if self._query is None:
            self._query = QueryList(self)
        return self._query

    @property
    def rate_limits(self) -> RateLimitList:
        if self._rate_limits is None:
            self._rate_limits = RateLimitList(self)
        return self._rate_limits

    def __repr__(self) -> str:
        """
        Provide a friendly representation
        :returns: Machine friendly representation
        """
        return "<Twilio.Lookups.V2>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Chat
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from datetime import datetime
from typing import Any, Dict, List, Optional, Union, Iterator, AsyncIterator
from twilio.base import deserialize, serialize, values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version
from twilio.base.page import Page


class MemberInstance(InstanceResource):
    """
    :ivar sid: The unique string that we created to identify the Member resource.
    :ivar account_sid: The SID of the [Account](https://www.twilio.com/docs/api/rest/account) that created the Member resource.
    :ivar channel_sid: The unique ID of the [Channel](https://www.twilio.com/docs/api/chat/rest/channels) for the member.
    :ivar service_sid: The SID of the [Service](https://www.twilio.com/docs/api/chat/rest/services) the resource is associated with.
    :ivar identity: The application-defined string that uniquely identifies the resource's [User](https://www.twilio.com/docs/api/chat/rest/users) within the [Service](https://www.twilio.com/docs/api/chat/rest/services). See [access tokens](https://www.twilio.com/docs/api/chat/guides/create-tokens) for more info.
    :ivar date_created: The date and time in GMT when the resource was created specified in [RFC 2822](http://www.ietf.org/rfc/rfc2822.txt) format.
    :ivar date_updated: The date and time in GMT when the resource was last updated specified in [RFC 2822](http://www.ietf.org/rfc/rfc2822.txt) format.
    :ivar role_sid: The SID of the [Role](https://www.twilio.com/docs/api/chat/rest/roles) assigned to the member.
    :ivar last_consumed_message_index: The index of the last [Message](https://www.twilio.com/docs/api/chat/rest/messages) in the [Channel](https://www.twilio.com/docs/api/chat/rest/channels) that the Member has read.
    :ivar last_consumption_timestamp: The ISO 8601 timestamp string that represents the date-time of the last [Message](https://www.twilio.com/docs/api/chat/rest/messages) read event for the Member within the [Channel](https://www.twilio.com/docs/api/chat/rest/channels).
    :ivar url: The absolute URL of the Member resource.
    """

    def __init__(
        self,
        version: Version,
        payload: Dict[str, Any],
        service_sid: str,
        channel_sid: str,
        sid: Optional[str] = None,
    ):
        super().__init__(version)

        self.sid: Optional[str] = payload.get("sid")
        self.account_sid: Optional[str] = payload.get("account_sid")
        self.channel_sid: Optional[str] = payload.get("channel_sid")
        self.service_sid: Optional[str] = payload.get("service_sid")
        self.identity: Optional[str] = payload.get("identity")
        self.date_created: Optional[datetime] = deserialize.iso8601_datetime(
            payload.get("date_created")
        )
        self.date_updated: Optional[datetime] = deserialize.iso8601_datetime(
            payload.get("date_updated")
        )
        self.role_sid: Optional[str] = payload.get("role_sid")
        self.last_consumed_message_index: Optional[int] = deserialize.integer(
            payload.get("last_consumed_message_index")
        )
        self.last_consumption_timestamp: Optional[datetime] = (
            deserialize.iso8601_datetime(payload.get("last_consumption_timestamp"))
        )
        self.url: Optional[str] = payload.get("url")

        self._solution = {
            "service_sid": service_sid,
            "channel_sid": channel_sid,
            "sid": sid or self.sid,
        }
        self._context: Optional[MemberContext] = None

    @property
    def _proxy(self) -> "MemberContext":
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: MemberContext for this MemberInstance
        """
        if self._context is None:
            self._context = MemberContext(
                self._version,
                service_sid=self._solution["service_sid"],
                channel_sid=self._solution["channel_sid"],
                sid=self._solution["sid"],
            )
        return self._context

    def delete(self) -> bool:
        """
        Deletes the MemberInstance


        :returns: True if delete succeeds, False otherwise
        """
        return self._proxy.delete()

    async def delete_async(self) -> bool:
        """
        Asynchronous coroutine that deletes the MemberInstance


        :returns: True if delete succeeds, False otherwise
        """
        return await self._proxy.delete_async()

    def fetch(self) -> "MemberInstance":
        """
        Fetch the MemberInstance


        :returns: The fetched MemberInstance
        """
        return self._proxy.fetch()

    async def fetch_async(self) -> "MemberInstance":
        """
        Asynchronous coroutine to fetch the MemberInstance


        :returns: The fetched MemberInstance
        """
        return await self._proxy.fetch_async()

    def update(
        self,
        role_sid: Union[str, object] = values.unset,
        last_consumed_message_index: Union[int, object] = values.unset,
    ) -> "MemberInstance":
        """
        Update the MemberInstance

        :param role_sid: The SID of the [Role](https://www.twilio.com/docs/api/chat/rest/roles) to assign to the member. The default roles are those specified on the [Service](https://www.twilio.com/docs/chat/api/services).
        :param last_consumed_message_index: The index of the last [Message](https://www.twilio.com/docs/api/chat/rest/messages) that the Member has read within the [Channel](https://www.twilio.com/docs/api/chat/rest/channels).

        :returns: The updated MemberInstance
        """
        return self._proxy.update(
            role_sid=role_sid,
            last_consumed_message_index=last_consumed_message_index,
        )

    async def update_async(
        self,
        role_sid: Union[str, object] = values.unset,
        last_consumed_message_index: Union[int, object] = values.unset,
    ) -> "MemberInstance":
        """
        Asynchronous coroutine to update the MemberInstance

        :param role_sid: The SID of the [Role](https://www.twilio.com/docs/api/chat/rest/roles) to assign to the member. The default roles are those specified on the [Service](https://www.twilio.com/docs/chat/api/services).
        :param last_consumed_message_index: The index of the last [Message](https://www.twilio.com/docs/api/chat/rest/messages) that the Member has read within the [Channel](https://www.twilio.com/docs/api/chat/rest/channels).

        :returns: The updated MemberInstance
        """
        return await self._proxy.update_async(
            role_sid=role_sid,
            last_consumed_message_index=last_consumed_message_index,
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Chat.V1.MemberInstance {}>".format(context)


class MemberContext(InstanceContext):

    def __init__(self, version: Version, service_sid: str, channel_sid: str, sid: str):
        """
        Initialize the MemberContext

        :param version: Version that contains the resource
        :param service_sid: The SID of the [Service](https://www.twilio.com/docs/api/chat/rest/services) to update the resource from.
        :param channel_sid: The unique ID of the [Channel](https://www.twilio.com/docs/api/chat/rest/channels) the member to update belongs to. Can be the Channel resource's `sid` or `unique_name`.
        :param sid: The Twilio-provided string that uniquely identifies the Member resource to update.
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "service_sid": service_sid,
            "channel_sid": channel_sid,
            "sid": sid,
        }
        self._uri = (
            "/Services/{service_sid}/Channels/{channel_sid}/Members/{sid}".format(
                **self._solution
            )
        )

    def delete(self) -> bool:
        """
        Deletes the MemberInstance


        :returns: True if delete succeeds, False otherwise
        """

        headers = values.of({})

        return self._version.delete(method="DELETE", uri=self._uri, headers=headers)

    async def delete_async(self) -> bool:
        """
        Asynchronous coroutine that deletes the MemberInstance


        :returns: True if delete succeeds, False otherwise
        """

        headers = values.of({})

        return await self._version.delete_async(
            method="DELETE", uri=self._uri, headers=headers
        )

    def fetch(self) -> MemberInstance:
        """
        Fetch the MemberInstance


        :returns: The fetched MemberInstance
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = self._version.fetch(method="GET", uri=self._uri, headers=headers)

        return MemberInstance(
            self._version,
            payload,
            service_sid=self._solution["service_sid"],
            channel_sid=self._solution["channel_sid"],
            sid=self._solution["sid"],
        )

    async def fetch_async(self) -> MemberInstance:
        """
        Asynchronous coroutine to fetch the MemberInstance


        :returns: The fetched MemberInstance
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = await self._version.fetch_async(
            method="GET", uri=self._uri, headers=headers
        )

        return MemberInstance(
            self._version,
            payload,
            service_sid=self._solution["service_sid"],
            channel_sid=self._solution["channel_sid"],
            sid=self._solution["sid"],
        )

    def update(
        self,
        role_sid: Union[str, object] = values.unset,
        last_consumed_message_index: Union[int, object] = values.unset,
    ) -> MemberInstance:
        """
        Update the MemberInstance

        :param role_sid: The SID of the [Role](https://www.twilio.com/docs/api/chat/rest/roles) to assign to the member. The default roles are those specified on the [Service](https://www.twilio.com/docs/chat/api/services).
        :param last_consumed_message_index: The index of the last [Message](https://www.twilio.com/docs/api/chat/rest/messages) that the Member has read within the [Channel](https://www.twilio.com/docs/api/chat/rest/channels).

        :returns: The updated MemberInstance
        """

        data = values.of(
            {
                "RoleSid": role_sid,
                "LastConsumedMessageIndex": last_consumed_message_index,
            }
        )
        headers = values.of({})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = self._version.update(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return MemberInstance(
            self._version,
            payload,
            service_sid=self._solution["service_sid"],
            channel_sid=self._solution["channel_sid"],
            sid=self._solution["sid"],
        )

    async def update_async(
        self,
        role_sid: Union[str, object] = values.unset,
        last_consumed_message_index: Union[int, object] = values.unset,
    ) -> MemberInstance:
        """
        Asynchronous coroutine to update the MemberInstance

        :param role_sid: The SID of the [Role](https://www.twilio.com/docs/api/chat/rest/roles) to assign to the member. The default roles are those specified on the [Service](https://www.twilio.com/docs/chat/api/services).
        :param last_consumed_message_index: The index of the last [Message](https://www.twilio.com/docs/api/chat/rest/messages) that the Member has read within the [Channel](https://www.twilio.com/docs/api/chat/rest/channels).

        :returns: The updated MemberInstance
        """

        data = values.of(
            {
                "RoleSid": role_sid,
                "LastConsumedMessageIndex": last_consumed_message_index,
            }
        )
        headers = values.of({})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = await self._version.update_async(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return MemberInstance(
            self._version,
            payload,
            service_sid=self._solution["service_sid"],
            channel_sid=self._solution["channel_sid"],
            sid=self._solution["sid"],
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Chat.V1.MemberContext {}>".format(context)


class MemberPage(Page):

    def get_instance(self, payload: Dict[str, Any]) -> MemberInstance:
        """
        Build an instance of MemberInstance

        :param payload: Payload response from the API
        """
        return MemberInstance(
            self._version,
            payload,
            service_sid=self._solution["service_sid"],
            channel_sid=self._solution["channel_sid"],
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Chat.V1.MemberPage>"


class MemberList(ListResource):

    def __init__(self, version: Version, service_sid: str, channel_sid: str):
        """
        Initialize the MemberList

        :param version: Version that contains the resource
        :param service_sid: The SID of the [Service](https://www.twilio.com/docs/api/chat/rest/services) to read the resources from.
        :param channel_sid: The unique ID of the [Channel](https://www.twilio.com/docs/api/chat/rest/channels) the members to read belong to. Can be the Channel resource's `sid` or `unique_name` value.

        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "service_sid": service_sid,
            "channel_sid": channel_sid,
        }
        self._uri = "/Services/{service_sid}/Channels/{channel_sid}/Members".format(
            **self._solution
        )

    def create(
        self, identity: str, role_sid: Union[str, object] = values.unset
    ) -> MemberInstance:
        """
        Create the MemberInstance

        :param identity: The `identity` value that uniquely identifies the new resource's [User](https://www.twilio.com/docs/api/chat/rest/v1/user) within the [Service](https://www.twilio.com/docs/api/chat/rest/services). See [access tokens](https://www.twilio.com/docs/api/chat/guides/create-tokens) for more details.
        :param role_sid: The SID of the [Role](https://www.twilio.com/docs/api/chat/rest/roles) to assign to the member. The default roles are those specified on the [Service](https://www.twilio.com/docs/chat/api/services).

        :returns: The created MemberInstance
        """

        data = values.of(
            {
                "Identity": identity,
                "RoleSid": role_sid,
            }
        )
        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = self._version.create(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return MemberInstance(
            self._version,
            payload,
            service_sid=self._solution["service_sid"],
            channel_sid=self._solution["channel_sid"],
        )

    async def create_async(
        self, identity: str, role_sid: Union[str, object] = values.unset
    ) -> MemberInstance:
        """
        Asynchronously create the MemberInstance

        :param identity: The `identity` value that uniquely identifies the new resource's [User](https://www.twilio.com/docs/api/chat/rest/v1/user) within the [Service](https://www.twilio.com/docs/api/chat/rest/services). See [access tokens](https://www.twilio.com/docs/api/chat/guides/create-tokens) for more details.
        :param role_sid: The SID of the [Role](https://www.twilio.com/docs/api/chat/rest/roles) to assign to the member. The default roles are those specified on the [Service](https://www.twilio.com/docs/chat/api/services).

        :returns: The created MemberInstance
        """

        data = values.of(
            {
                "Identity": identity,
                "RoleSid": role_sid,
            }
        )
        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = await self._version.create_async(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return MemberInstance(
            self._version,
            payload,
            service_sid=self._solution["service_sid"],
            channel_sid=self._solution["channel_sid"],
        )

    def stream(
        self,
        identity: Union[List[str], object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Iterator[MemberInstance]:
        """
        Streams MemberInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param List[str] identity: The [User](https://www.twilio.com/docs/api/chat/rest/v1/user)'s `identity` value of the resources to read. See [access tokens](https://www.twilio.com/docs/api/chat/guides/create-tokens) for more details.
        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = self.page(identity=identity, page_size=limits["page_size"])

        return self._version.stream(page, limits["limit"])

    async def stream_async(
        self,
        identity: Union[List[str], object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> AsyncIterator[MemberInstance]:
        """
        Asynchronously streams MemberInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param List[str] identity: The [User](https://www.twilio.com/docs/api/chat/rest/v1/user)'s `identity` value of the resources to read. See [access tokens](https://www.twilio.com/docs/api/chat/guides/create-tokens) for more details.
        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = await self.page_async(identity=identity, page_size=limits["page_size"])

        return self._version.stream_async(page, limits["limit"])

    def list(
        self,
        identity: Union[List[str], object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[MemberInstance]:
        """
        Lists MemberInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param List[str] identity: The [User](https://www.twilio.com/docs/api/chat/rest/v1/user)'s `identity` value of the resources to read. See [access tokens](https://www.twilio.com/docs/api/chat/guides/create-tokens) for more details.
        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return list(
            self.stream(
                identity=identity,
                limit=limit,
                page_size=page_size,
            )
        )

    async def list_async(
        self,
        identity: Union[List[str], object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[MemberInstance]:
        """
        Asynchronously lists MemberInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param List[str] identity: The [User](https://www.twilio.com/docs/api/chat/rest/v1/user)'s `identity` value of the resources to read. See [access tokens](https://www.twilio.com/docs/api/chat/guides/create-tokens) for more details.
        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return [
            record
            async for record in await self.stream_async(
                identity=identity,
                limit=limit,
                page_size=page_size,
            )
        ]

    def page(
        self,
        identity: Union[List[str], object] = values.unset,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> MemberPage:
        """
        Retrieve a single page of MemberInstance records from the API.
        Request is executed immediately

        :param identity: The [User](https://www.twilio.com/docs/api/chat/rest/v1/user)'s `identity` value of the resources to read. See [access tokens](https://www.twilio.com/docs/api/chat/guides/create-tokens) for more details.
        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of MemberInstance
        """
        data = values.of(
            {
                "Identity": serialize.map(identity, lambda e: e),
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response = self._version.page(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return MemberPage(self._version, response, self._solution)

    async def page_async(
        self,
        identity: Union[List[str], object] = values.unset,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> MemberPage:
        """
        Asynchronously retrieve a single page of MemberInstance records from the API.
        Request is executed immediately

        :param identity: The [User](https://www.twilio.com/docs/api/chat/rest/v1/user)'s `identity` value of the resources to read. See [access tokens](https://www.twilio.com/docs/api/chat/guides/create-tokens) for more details.
        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of MemberInstance
        """
        data = values.of(
            {
                "Identity": serialize.map(identity, lambda e: e),
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response = await self._version.page_async(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return MemberPage(self._version, response, self._solution)

    def get_page(self, target_url: str) -> MemberPage:
        """
        Retrieve a specific page of MemberInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of MemberInstance
        """
        response = self._version.domain.twilio.request("GET", target_url)
        return MemberPage(self._version, response, self._solution)

    async def get_page_async(self, target_url: str) -> MemberPage:
        """
        Asynchronously retrieve a specific page of MemberInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of MemberInstance
        """
        response = await self._version.domain.twilio.request_async("GET", target_url)
        return MemberPage(self._version, response, self._solution)

    def get(self, sid: str) -> MemberContext:
        """
        Constructs a MemberContext

        :param sid: The Twilio-provided string that uniquely identifies the Member resource to update.
        """
        return MemberContext(
            self._version,
            service_sid=self._solution["service_sid"],
            channel_sid=self._solution["channel_sid"],
            sid=sid,
        )

    def __call__(self, sid: str) -> MemberContext:
        """
        Constructs a MemberContext

        :param sid: The Twilio-provided string that uniquely identifies the Member resource to update.
        """
        return MemberContext(
            self._version,
            service_sid=self._solution["service_sid"],
            channel_sid=self._solution["channel_sid"],
            sid=sid,
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Chat.V1.MemberList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Verify
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from datetime import datetime
from typing import Any, Dict, Optional
from twilio.base import deserialize, values

from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version


class ApproveChallengeInstance(InstanceResource):

    class ApprovePasskeysChallengeRequest(object):
        """
        :ivar id: A [base64url](https://base64.guru/standards/base64url) encoded representation of `rawId`.
        :ivar raw_id: The globally unique identifier for this `PublicKeyCredential`.
        :ivar authenticator_attachment: A string that indicates the mechanism by which the WebAuthn implementation is attached to the authenticator at the time the associated `navigator.credentials.create()` or `navigator.credentials.get()` call completes.
        :ivar type: The valid credential types supported by the API. The values of this enumeration are used for versioning the `AuthenticatorAssertion` and `AuthenticatorAttestation` structures according to the type of the authenticator.
        :ivar response:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.id: Optional[str] = payload.get("id")
            self.raw_id: Optional[str] = payload.get("raw_id")
            self.authenticator_attachment: Optional["ApproveChallengeInstance.str"] = (
                payload.get("authenticator_attachment")
            )
            self.type: Optional["ApproveChallengeInstance.str"] = payload.get("type")
            self.response: Optional[
                ApproveChallengeList.ApprovePasskeysChallengeRequestResponse
            ] = payload.get("response")

        def to_dict(self):
            return {
                "id": self.id,
                "raw_id": self.raw_id,
                "authenticator_attachment": self.authenticator_attachment,
                "type": self.type,
                "response": (
                    self.response.to_dict() if self.response is not None else None
                ),
            }

    class ApprovePasskeysChallengeRequestResponse(object):
        """
        :ivar authenticator_data: The [authenticator data](https://developer.mozilla.org/en-US/docs/Web/API/Web_Authentication_API/Authenticator_data) structure contains information from the authenticator about the processing of a credential creation or authentication request.
        :ivar client_data_json: This property contains the JSON-compatible serialization of the data passed from the browser to the authenticator in order to generate this credential.
        :ivar signature: An assertion signature over `authenticatorData` and `clientDataJSON`. The assertion signature is created with the private key of the key pair that was created during the originating `navigator.credentials.create()` call and verified using the public key of that same key pair.
        :ivar user_handle: The user handle stored in the authenticator, specified as `user.id` in the options passed to the originating `navigator.credentials.create()` call. This property should contain a base64url-encoded entity SID.
        """

        def __init__(self, payload: Dict[str, Any]):

            self.authenticator_data: Optional[str] = payload.get("authenticator_data")
            self.client_data_json: Optional[str] = payload.get("client_data_json")
            self.signature: Optional[str] = payload.get("signature")
            self.user_handle: Optional[str] = payload.get("user_handle")

        def to_dict(self):
            return {
                "authenticator_data": self.authenticator_data,
                "client_data_json": self.client_data_json,
                "signature": self.signature,
                "user_handle": self.user_handle,
            }

    """
    :ivar options: An object that contains challenge options. Currently only used for `passkeys`.
    :ivar sid: A 34 character string that uniquely identifies this Challenge.
    :ivar account_sid: The unique SID identifier of the Account.
    :ivar service_sid: The unique SID identifier of the Service.
    :ivar entity_sid: The unique SID identifier of the Entity.
    :ivar identity: Customer unique identity for the Entity owner of the Challenge.
    :ivar factor_sid: The unique SID identifier of the Factor.
    :ivar date_created: The date that this Challenge was created, given in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
    :ivar date_updated: The date that this Challenge was updated, given in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
    :ivar date_responded: The date that this Challenge was responded, given in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
    :ivar expiration_date: The date-time when this Challenge expires, given in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
    :ivar status: The Status of this Challenge. One of `pending`, `expired`, `approved` or `denied`.
    :ivar responded_reason: Reason for the Challenge to be in certain `status`. One of `none`, `not_needed` or `not_requested`.
    :ivar details: Details provided to give context about the Challenge.
    :ivar hidden_details: Details provided to give context about the Challenge.
    :ivar metadata: Custom metadata associated with the challenge.
    :ivar factor_type: The Factor Type of this Challenge. Currently `push` and `totp` are supported.
    :ivar url: The URL of this resource.
    :ivar links: Contains a dictionary of URL links to nested resources of this Challenge.
    """

    def __init__(self, version: Version, payload: Dict[str, Any], service_sid: str):
        super().__init__(version)

        self.options: Optional[Dict[str, object]] = payload.get("options")
        self.sid: Optional[str] = payload.get("sid")
        self.account_sid: Optional[str] = payload.get("account_sid")
        self.service_sid: Optional[str] = payload.get("service_sid")
        self.entity_sid: Optional[str] = payload.get("entity_sid")
        self.identity: Optional[str] = payload.get("identity")
        self.factor_sid: Optional[str] = payload.get("factor_sid")
        self.date_created: Optional[datetime] = deserialize.iso8601_datetime(
            payload.get("date_created")
        )
        self.date_updated: Optional[datetime] = deserialize.iso8601_datetime(
            payload.get("date_updated")
        )
        self.date_responded: Optional[datetime] = deserialize.iso8601_datetime(
            payload.get("date_responded")
        )
        self.expiration_date: Optional[datetime] = deserialize.iso8601_datetime(
            payload.get("expiration_date")
        )
        self.status: Optional[str] = payload.get("status")
        self.responded_reason: Optional[str] = payload.get("responded_reason")
        self.details: Optional[Dict[str, object]] = payload.get("details")
        self.hidden_details: Optional[Dict[str, object]] = payload.get("hidden_details")
        self.metadata: Optional[Dict[str, object]] = payload.get("metadata")
        self.factor_type: Optional[str] = payload.get("factor_type")
        self.url: Optional[str] = payload.get("url")
        self.links: Optional[Dict[str, object]] = payload.get("links")

        self._solution = {
            "service_sid": service_sid,
        }

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Verify.V2.ApproveChallengeInstance {}>".format(context)


class ApproveChallengeList(ListResource):

    class ApprovePasskeysChallengeRequest(object):
        """
        :ivar id: A [base64url](https://base64.guru/standards/base64url) encoded representation of `rawId`.
        :ivar raw_id: The globally unique identifier for this `PublicKeyCredential`.
        :ivar authenticator_attachment: A string that indicates the mechanism by which the WebAuthn implementation is attached to the authenticator at the time the associated `navigator.credentials.create()` or `navigator.credentials.get()` call completes.
        :ivar type: The valid credential types supported by the API. The values of this enumeration are used for versioning the `AuthenticatorAssertion` and `AuthenticatorAttestation` structures according to the type of the authenticator.
        :ivar response:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.id: Optional[str] = payload.get("id")
            self.raw_id: Optional[str] = payload.get("raw_id")
            self.authenticator_attachment: Optional["ApproveChallengeInstance.str"] = (
                payload.get("authenticator_attachment")
            )
            self.type: Optional["ApproveChallengeInstance.str"] = payload.get("type")
            self.response: Optional[
                ApproveChallengeList.ApprovePasskeysChallengeRequestResponse
            ] = payload.get("response")

        def to_dict(self):
            return {
                "id": self.id,
                "raw_id": self.raw_id,
                "authenticator_attachment": self.authenticator_attachment,
                "type": self.type,
                "response": (
                    self.response.to_dict() if self.response is not None else None
                ),
            }

    class ApprovePasskeysChallengeRequestResponse(object):
        """
        :ivar authenticator_data: The [authenticator data](https://developer.mozilla.org/en-US/docs/Web/API/Web_Authentication_API/Authenticator_data) structure contains information from the authenticator about the processing of a credential creation or authentication request.
        :ivar client_data_json: This property contains the JSON-compatible serialization of the data passed from the browser to the authenticator in order to generate this credential.
        :ivar signature: An assertion signature over `authenticatorData` and `clientDataJSON`. The assertion signature is created with the private key of the key pair that was created during the originating `navigator.credentials.create()` call and verified using the public key of that same key pair.
        :ivar user_handle: The user handle stored in the authenticator, specified as `user.id` in the options passed to the originating `navigator.credentials.create()` call. This property should contain a base64url-encoded entity SID.
        """

        def __init__(self, payload: Dict[str, Any]):

            self.authenticator_data: Optional[str] = payload.get("authenticator_data")
            self.client_data_json: Optional[str] = payload.get("client_data_json")
            self.signature: Optional[str] = payload.get("signature")
            self.user_handle: Optional[str] = payload.get("user_handle")

        def to_dict(self):
            return {
                "authenticator_data": self.authenticator_data,
                "client_data_json": self.client_data_json,
                "signature": self.signature,
                "user_handle": self.user_handle,
            }

    def __init__(self, version: Version, service_sid: str):
        """
        Initialize the ApproveChallengeList

        :param version: Version that contains the resource
        :param service_sid: The unique SID identifier of the Service.

        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "service_sid": service_sid,
        }
        self._uri = "/Services/{service_sid}/Passkeys/ApproveChallenge".format(
            **self._solution
        )

    def update(
        self, approve_passkeys_challenge_request: ApprovePasskeysChallengeRequest
    ) -> ApproveChallengeInstance:
        """
        Update the ApproveChallengeInstance

        :param approve_passkeys_challenge_request:

        :returns: The created ApproveChallengeInstance
        """
        data = approve_passkeys_challenge_request.to_dict()

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/json"

        headers["Accept"] = "application/json"

        payload = self._version.update(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return ApproveChallengeInstance(
            self._version, payload, service_sid=self._solution["service_sid"]
        )

    async def update_async(
        self, approve_passkeys_challenge_request: ApprovePasskeysChallengeRequest
    ) -> ApproveChallengeInstance:
        """
        Asynchronously update the ApproveChallengeInstance

        :param approve_passkeys_challenge_request:

        :returns: The created ApproveChallengeInstance
        """
        data = approve_passkeys_challenge_request.to_dict()

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/json"

        headers["Accept"] = "application/json"

        payload = await self._version.update_async(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return ApproveChallengeInstance(
            self._version, payload, service_sid=self._solution["service_sid"]
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Verify.V2.ApproveChallengeList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Accounts
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from typing import Any, Dict, List, Optional, Union
from twilio.base import serialize, values

from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version


class MessagingGeopermissionsInstance(InstanceResource):
    """
    :ivar permissions: A list of objects where each object represents the result of processing a messaging Geo Permission. Each object contains the following fields: `country_code`, the country code of the country for which the permission was updated; `type`, the type of the permission i.e. country; `enabled`, true if the permission is enabled else false; `error_code`, an integer where 0 indicates success and any non-zero value represents an error; and `error_messages`, an array of strings describing specific validation errors encountered. If the request is successful, the error_messages array will be empty.
    """

    def __init__(self, version: Version, payload: Dict[str, Any]):
        super().__init__(version)

        self.permissions: Optional[Dict[str, object]] = payload.get("permissions")

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """

        return "<Twilio.Accounts.V1.MessagingGeopermissionsInstance>"


class MessagingGeopermissionsList(ListResource):

    def __init__(self, version: Version):
        """
        Initialize the MessagingGeopermissionsList

        :param version: Version that contains the resource

        """
        super().__init__(version)

        self._uri = "/Messaging/GeoPermissions"

    def fetch(
        self, country_code: Union[str, object] = values.unset
    ) -> MessagingGeopermissionsInstance:
        """
        Asynchronously fetch the MessagingGeopermissionsInstance

        :param country_code: The country code to filter the geo permissions. If provided, only the geo permission for the specified country will be returned.
        :returns: The fetched MessagingGeopermissionsInstance
        """
        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        params = values.of(
            {
                "CountryCode": country_code,
            }
        )

        payload = self._version.fetch(
            method="GET", uri=self._uri, headers=headers, params=params
        )

        return MessagingGeopermissionsInstance(self._version, payload)

    async def fetch_async(
        self, country_code: Union[str, object] = values.unset
    ) -> MessagingGeopermissionsInstance:
        """
        Asynchronously fetch the MessagingGeopermissionsInstance

        :param country_code: The country code to filter the geo permissions. If provided, only the geo permission for the specified country will be returned.
        :returns: The fetched MessagingGeopermissionsInstance
        """
        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        params = values.of(
            {
                "CountryCode": country_code,
            }
        )

        payload = await self._version.fetch_async(
            method="GET", uri=self._uri, headers=headers, params=params
        )

        return MessagingGeopermissionsInstance(self._version, payload)

    def update(self, permissions: List[object]) -> MessagingGeopermissionsInstance:
        """
        Update the MessagingGeopermissionsInstance

        :param permissions: A list of objects where each object represents the Geo Permission to be updated. Each object contains the following fields: `country_code`, unique code for each country of Geo Permission; `type`, permission type of the Geo Permission i.e. country; `enabled`, configure true for enabling the Geo Permission, false for disabling the Geo Permission.

        :returns: The created MessagingGeopermissionsInstance
        """

        data = values.of(
            {
                "Permissions": serialize.map(
                    permissions, lambda e: serialize.object(e)
                ),
            }
        )
        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = self._version.update(
            method="PATCH", uri=self._uri, data=data, headers=headers
        )

        return MessagingGeopermissionsInstance(self._version, payload)

    async def update_async(
        self, permissions: List[object]
    ) -> MessagingGeopermissionsInstance:
        """
        Asynchronously update the MessagingGeopermissionsInstance

        :param permissions: A list of objects where each object represents the Geo Permission to be updated. Each object contains the following fields: `country_code`, unique code for each country of Geo Permission; `type`, permission type of the Geo Permission i.e. country; `enabled`, configure true for enabling the Geo Permission, false for disabling the Geo Permission.

        :returns: The created MessagingGeopermissionsInstance
        """

        data = values.of(
            {
                "Permissions": serialize.map(
                    permissions, lambda e: serialize.object(e)
                ),
            }
        )
        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = await self._version.update_async(
            method="PATCH", uri=self._uri, data=data, headers=headers
        )

        return MessagingGeopermissionsInstance(self._version, payload)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Accounts.V1.MessagingGeopermissionsList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Numbers
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from datetime import date, datetime
from typing import Any, Dict, List, Optional
from twilio.base import deserialize, values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version


class PortingPortInInstance(InstanceResource):

    class NumbersV1PortingAddress(object):
        """
        :ivar street: The street address, ex: 101 Spear St
        :ivar street_2: The building information, ex : 5th floor.
        :ivar city: The city name, ex: San Francisco.
        :ivar state: The state name, ex: CA or California. Note this should match the losing carrier’s information exactly. So if they spell out the entire state’s name instead of abbreviating it, please do so.
        :ivar zip: The zip code, ex: 94105.
        :ivar country: The country, ex: USA.
        """

        def __init__(self, payload: Dict[str, Any]):

            self.street: Optional[str] = payload.get("street")
            self.street_2: Optional[str] = payload.get("street_2")
            self.city: Optional[str] = payload.get("city")
            self.state: Optional[str] = payload.get("state")
            self.zip: Optional[str] = payload.get("zip")
            self.country: Optional[str] = payload.get("country")

        def to_dict(self):
            return {
                "street": self.street,
                "street_2": self.street_2,
                "city": self.city,
                "state": self.state,
                "zip": self.zip,
                "country": self.country,
            }

    class NumbersV1PortingLosingCarrierInformation(object):
        """
        :ivar customer_name: Customer name as it is registered with the losing carrier. This can be an individual or a business name depending on the customer type selected.
        :ivar account_number: The account number of the customer for the losing carrier. Only require for mobile phone numbers.
        :ivar account_telephone_number: The account phone number of the customer for the losing carrier.
        :ivar address_sid: If you already have an Address SID that represents the address needed for the LOA, you can provide an Address SID instead of providing the address object in the request body. This will copy the address into the port in request. If changes are made to the Address SID after port in request creation, those changes will not be reflected in the port in request.
        :ivar address:
        :ivar authorized_representative: The first and last name of the person listed with the losing carrier who is authorized to make changes on the account.
        :ivar authorized_representative_email: Email address of the person (owner of the number) who will sign the letter of authorization for the port in request. This email address should belong to the person named in as the authorized representative.
        :ivar customer_type: The type of customer account in the losing carrier. This should either be: 'Individual' or 'Business'.
        :ivar authorized_representative_katakana:
        :ivar sub_municipality:
        :ivar building:
        :ivar katakana_name:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.customer_name: Optional[str] = payload.get("customer_name")
            self.account_number: Optional[str] = payload.get("account_number")
            self.account_telephone_number: Optional[str] = payload.get(
                "account_telephone_number"
            )
            self.address_sid: Optional[str] = payload.get("address_sid")
            self.address: Optional[PortingPortInList.NumbersV1PortingAddress] = (
                payload.get("address")
            )
            self.authorized_representative: Optional[str] = payload.get(
                "authorized_representative"
            )
            self.authorized_representative_email: Optional[str] = payload.get(
                "authorized_representative_email"
            )
            self.customer_type: Optional["PortingPortInInstance.str"] = payload.get(
                "customer_type"
            )
            self.authorized_representative_katakana: Optional[str] = payload.get(
                "authorized_representative_katakana"
            )
            self.sub_municipality: Optional[str] = payload.get("sub_municipality")
            self.building: Optional[str] = payload.get("building")
            self.katakana_name: Optional[str] = payload.get("katakana_name")

        def to_dict(self):
            return {
                "customer_name": self.customer_name,
                "account_number": self.account_number,
                "account_telephone_number": self.account_telephone_number,
                "address_sid": self.address_sid,
                "address": self.address.to_dict() if self.address is not None else None,
                "authorized_representative": self.authorized_representative,
                "authorized_representative_email": self.authorized_representative_email,
                "customer_type": self.customer_type,
                "authorized_representative_katakana": self.authorized_representative_katakana,
                "sub_municipality": self.sub_municipality,
                "building": self.building,
                "katakana_name": self.katakana_name,
            }

    class NumbersV1PortingPortInCreate(object):
        """
        :ivar account_sid: Account Sid or subaccount where the phone number(s) will be Ported
        :ivar documents: List of document SIDs for all phone numbers included in the port in request. At least one document SID referring to a document of the type Utility Bill is required.
        :ivar phone_numbers: List of phone numbers to be ported. Maximum of 1,000 phone numbers per request.
        :ivar losing_carrier_information:
        :ivar notification_emails: Additional emails to send a copy of the signed LOA to.
        :ivar target_port_in_date: Target date to port the number. We cannot guarantee that this date will be honored by the other carriers, please work with Ops to get a confirmation of the firm order commitment (FOC) date. Expected format is ISO Local Date, example: ‘2011-12-03`. This date must be at least 7 days in the future for US ports and 10 days in the future for Japanese ports. We can't guarantee the exact date and time, as this depends on the losing carrier
        :ivar target_port_in_time_range_start: The earliest time that the port should occur on the target port in date. Expected format is ISO Offset Time, example: ‘10:15:00-08:00'. We can't guarantee the exact date and time, as this depends on the losing carrier
        :ivar target_port_in_time_range_end: The latest time that the port should occur on the target port in date. Expected format is ISO Offset Time, example: ‘10:15:00-08:00'. We can't guarantee the exact date and time, as this depends on the losing carrier
        :ivar bundle_sid: The bundle sid is an optional identifier to reference a group of regulatory documents for a port request.
        :ivar portability_advance_carrier: A field only required for Japan port in requests. It is a unique identifier for the donor carrier service the line is being ported from.
        :ivar auto_cancel_approval_numbers: Japan specific field, indicates the number of phone numbers to automatically approve for cancellation.
        """

        def __init__(self, payload: Dict[str, Any]):

            self.account_sid: Optional[str] = payload.get("account_sid")
            self.documents: Optional[List[str]] = payload.get("documents")
            self.phone_numbers: Optional[
                List[PortingPortInList.NumbersV1PortingPortInCreatePhoneNumbers]
            ] = payload.get("phone_numbers")
            self.losing_carrier_information: Optional[
                PortingPortInList.NumbersV1PortingLosingCarrierInformation
            ] = payload.get("losing_carrier_information")
            self.notification_emails: Optional[List[str]] = payload.get(
                "notification_emails"
            )
            self.target_port_in_date: Optional[date] = payload.get(
                "target_port_in_date"
            )
            self.target_port_in_time_range_start: Optional[str] = payload.get(
                "target_port_in_time_range_start"
            )
            self.target_port_in_time_range_end: Optional[str] = payload.get(
                "target_port_in_time_range_end"
            )
            self.bundle_sid: Optional[str] = payload.get("bundle_sid")
            self.portability_advance_carrier: Optional[str] = payload.get(
                "portability_advance_carrier"
            )
            self.auto_cancel_approval_numbers: Optional[str] = payload.get(
                "auto_cancel_approval_numbers"
            )

        def to_dict(self):
            return {
                "account_sid": self.account_sid,
                "documents": self.documents,
                "phone_numbers": (
                    [phone_numbers.to_dict() for phone_numbers in self.phone_numbers]
                    if self.phone_numbers is not None
                    else None
                ),
                "losing_carrier_information": (
                    self.losing_carrier_information.to_dict()
                    if self.losing_carrier_information is not None
                    else None
                ),
                "notification_emails": self.notification_emails,
                "target_port_in_date": self.target_port_in_date,
                "target_port_in_time_range_start": self.target_port_in_time_range_start,
                "target_port_in_time_range_end": self.target_port_in_time_range_end,
                "bundle_sid": self.bundle_sid,
                "portability_advance_carrier": self.portability_advance_carrier,
                "auto_cancel_approval_numbers": self.auto_cancel_approval_numbers,
            }

    class NumbersV1PortingPortInCreatePhoneNumbers(object):
        """
        :ivar phone_number: Phone number to be ported. This must be in the E164 Format.
        :ivar pin: Some losing carriers require a PIN to authorize the port of a phone number. If the phone number is a US mobile phone number, the PIN is mandatory to process a porting request. Other carriers and number types may also require a PIN, you'll need to contact the losing carrier to determine what your phone number's PIN is.
        """

        def __init__(self, payload: Dict[str, Any]):

            self.phone_number: Optional[str] = payload.get("phone_number")
            self.pin: Optional[str] = payload.get("pin")

        def to_dict(self):
            return {
                "phone_number": self.phone_number,
                "pin": self.pin,
            }

    """
    :ivar port_in_request_sid: The SID of the Port In request. This is a unique identifier of the port in request.
    :ivar url: The URL of this Port In request
    :ivar account_sid: Account Sid or subaccount where the phone number(s) will be Ported
    :ivar notification_emails: Additional emails to send a copy of the signed LOA to.
    :ivar target_port_in_date: Target date to port the number. We cannot guarantee that this date will be honored by the other carriers, please work with Ops to get a confirmation of the firm order commitment (FOC) date. Expected format is ISO Local Date, example: ‘2011-12-03`. This date must be at least 7 days in the future for US ports and 10 days in the future for Japanese ports. If a start and end range is provided, the date will be converted to its UTC equivalent with the ranges as reference and stored in UTC. We can't guarantee the exact date and time, as this depends on the losing carrier.
    :ivar target_port_in_time_range_start: The earliest time that the port should occur on the target port in date. Expected format is ISO Offset Time, example: ‘10:15:00-08:00'. We can't guarantee the exact date and time, as this depends on the losing carrier. The time will be stored and returned as UTC standard timezone.
    :ivar target_port_in_time_range_end: The latest time that the port should occur on the target port in date. Expected format is ISO Offset Time, example: ‘10:15:00-08:00'. We can't guarantee the exact date and time, as this depends on the losing carrier. The time will be stored and returned as UTC standard timezone.
    :ivar port_in_request_status: The status of the port in request. The possible values are: In progress, Completed, Expired, In review, Waiting for Signature, Action Required, and Canceled.
    :ivar order_cancellation_reason: If the order is cancelled this field will provide further context on the cause of the cancellation.
    :ivar losing_carrier_information: 
    :ivar phone_numbers: 
    :ivar bundle_sid: The bundle sid is an optional identifier to reference a group of regulatory documents for a port request.
    :ivar portability_advance_carrier: A field only required for Japan port in requests. It is a unique identifier for the donor carrier service the line is being ported from.
    :ivar auto_cancel_approval_numbers: Japan specific field, indicates the number of phone numbers to automatically approve for cancellation.
    :ivar documents: List of document SIDs for all phone numbers included in the port in request. At least one document SID referring to a document of the type Utility Bill is required.
    :ivar date_created: 
    :ivar support_ticket_id: Unique ID of the request's support ticket
    """

    def __init__(
        self,
        version: Version,
        payload: Dict[str, Any],
        port_in_request_sid: Optional[str] = None,
    ):
        super().__init__(version)

        self.port_in_request_sid: Optional[str] = payload.get("port_in_request_sid")
        self.url: Optional[str] = payload.get("url")
        self.account_sid: Optional[str] = payload.get("account_sid")
        self.notification_emails: Optional[List[str]] = payload.get(
            "notification_emails"
        )
        self.target_port_in_date: Optional[date] = deserialize.iso8601_date(
            payload.get("target_port_in_date")
        )
        self.target_port_in_time_range_start: Optional[str] = payload.get(
            "target_port_in_time_range_start"
        )
        self.target_port_in_time_range_end: Optional[str] = payload.get(
            "target_port_in_time_range_end"
        )
        self.port_in_request_status: Optional[str] = payload.get(
            "port_in_request_status"
        )
        self.order_cancellation_reason: Optional[str] = payload.get(
            "order_cancellation_reason"
        )
        self.losing_carrier_information: Optional[str] = payload.get(
            "losing_carrier_information"
        )
        self.phone_numbers: Optional[List[str]] = payload.get("phone_numbers")
        self.bundle_sid: Optional[str] = payload.get("bundle_sid")
        self.portability_advance_carrier: Optional[str] = payload.get(
            "portability_advance_carrier"
        )
        self.auto_cancel_approval_numbers: Optional[str] = payload.get(
            "auto_cancel_approval_numbers"
        )
        self.documents: Optional[List[str]] = payload.get("documents")
        self.date_created: Optional[datetime] = deserialize.iso8601_datetime(
            payload.get("date_created")
        )
        self.support_ticket_id: Optional[int] = deserialize.integer(
            payload.get("support_ticket_id")
        )

        self._solution = {
            "port_in_request_sid": port_in_request_sid or self.port_in_request_sid,
        }
        self._context: Optional[PortingPortInContext] = None

    @property
    def _proxy(self) -> "PortingPortInContext":
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: PortingPortInContext for this PortingPortInInstance
        """
        if self._context is None:
            self._context = PortingPortInContext(
                self._version,
                port_in_request_sid=self._solution["port_in_request_sid"],
            )
        return self._context

    def delete(self) -> bool:
        """
        Deletes the PortingPortInInstance


        :returns: True if delete succeeds, False otherwise
        """
        return self._proxy.delete()

    async def delete_async(self) -> bool:
        """
        Asynchronous coroutine that deletes the PortingPortInInstance


        :returns: True if delete succeeds, False otherwise
        """
        return await self._proxy.delete_async()

    def fetch(self) -> "PortingPortInInstance":
        """
        Fetch the PortingPortInInstance


        :returns: The fetched PortingPortInInstance
        """
        return self._proxy.fetch()

    async def fetch_async(self) -> "PortingPortInInstance":
        """
        Asynchronous coroutine to fetch the PortingPortInInstance


        :returns: The fetched PortingPortInInstance
        """
        return await self._proxy.fetch_async()

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Numbers.V1.PortingPortInInstance {}>".format(context)


class PortingPortInContext(InstanceContext):

    class NumbersV1PortingAddress(object):
        """
        :ivar street: The street address, ex: 101 Spear St
        :ivar street_2: The building information, ex : 5th floor.
        :ivar city: The city name, ex: San Francisco.
        :ivar state: The state name, ex: CA or California. Note this should match the losing carrier’s information exactly. So if they spell out the entire state’s name instead of abbreviating it, please do so.
        :ivar zip: The zip code, ex: 94105.
        :ivar country: The country, ex: USA.
        """

        def __init__(self, payload: Dict[str, Any]):

            self.street: Optional[str] = payload.get("street")
            self.street_2: Optional[str] = payload.get("street_2")
            self.city: Optional[str] = payload.get("city")
            self.state: Optional[str] = payload.get("state")
            self.zip: Optional[str] = payload.get("zip")
            self.country: Optional[str] = payload.get("country")

        def to_dict(self):
            return {
                "street": self.street,
                "street_2": self.street_2,
                "city": self.city,
                "state": self.state,
                "zip": self.zip,
                "country": self.country,
            }

    class NumbersV1PortingLosingCarrierInformation(object):
        """
        :ivar customer_name: Customer name as it is registered with the losing carrier. This can be an individual or a business name depending on the customer type selected.
        :ivar account_number: The account number of the customer for the losing carrier. Only require for mobile phone numbers.
        :ivar account_telephone_number: The account phone number of the customer for the losing carrier.
        :ivar address_sid: If you already have an Address SID that represents the address needed for the LOA, you can provide an Address SID instead of providing the address object in the request body. This will copy the address into the port in request. If changes are made to the Address SID after port in request creation, those changes will not be reflected in the port in request.
        :ivar address:
        :ivar authorized_representative: The first and last name of the person listed with the losing carrier who is authorized to make changes on the account.
        :ivar authorized_representative_email: Email address of the person (owner of the number) who will sign the letter of authorization for the port in request. This email address should belong to the person named in as the authorized representative.
        :ivar customer_type: The type of customer account in the losing carrier. This should either be: 'Individual' or 'Business'.
        :ivar authorized_representative_katakana:
        :ivar sub_municipality:
        :ivar building:
        :ivar katakana_name:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.customer_name: Optional[str] = payload.get("customer_name")
            self.account_number: Optional[str] = payload.get("account_number")
            self.account_telephone_number: Optional[str] = payload.get(
                "account_telephone_number"
            )
            self.address_sid: Optional[str] = payload.get("address_sid")
            self.address: Optional[PortingPortInList.NumbersV1PortingAddress] = (
                payload.get("address")
            )
            self.authorized_representative: Optional[str] = payload.get(
                "authorized_representative"
            )
            self.authorized_representative_email: Optional[str] = payload.get(
                "authorized_representative_email"
            )
            self.customer_type: Optional["PortingPortInInstance.str"] = payload.get(
                "customer_type"
            )
            self.authorized_representative_katakana: Optional[str] = payload.get(
                "authorized_representative_katakana"
            )
            self.sub_municipality: Optional[str] = payload.get("sub_municipality")
            self.building: Optional[str] = payload.get("building")
            self.katakana_name: Optional[str] = payload.get("katakana_name")

        def to_dict(self):
            return {
                "customer_name": self.customer_name,
                "account_number": self.account_number,
                "account_telephone_number": self.account_telephone_number,
                "address_sid": self.address_sid,
                "address": self.address.to_dict() if self.address is not None else None,
                "authorized_representative": self.authorized_representative,
                "authorized_representative_email": self.authorized_representative_email,
                "customer_type": self.customer_type,
                "authorized_representative_katakana": self.authorized_representative_katakana,
                "sub_municipality": self.sub_municipality,
                "building": self.building,
                "katakana_name": self.katakana_name,
            }

    class NumbersV1PortingPortInCreate(object):
        """
        :ivar account_sid: Account Sid or subaccount where the phone number(s) will be Ported
        :ivar documents: List of document SIDs for all phone numbers included in the port in request. At least one document SID referring to a document of the type Utility Bill is required.
        :ivar phone_numbers: List of phone numbers to be ported. Maximum of 1,000 phone numbers per request.
        :ivar losing_carrier_information:
        :ivar notification_emails: Additional emails to send a copy of the signed LOA to.
        :ivar target_port_in_date: Target date to port the number. We cannot guarantee that this date will be honored by the other carriers, please work with Ops to get a confirmation of the firm order commitment (FOC) date. Expected format is ISO Local Date, example: ‘2011-12-03`. This date must be at least 7 days in the future for US ports and 10 days in the future for Japanese ports. We can't guarantee the exact date and time, as this depends on the losing carrier
        :ivar target_port_in_time_range_start: The earliest time that the port should occur on the target port in date. Expected format is ISO Offset Time, example: ‘10:15:00-08:00'. We can't guarantee the exact date and time, as this depends on the losing carrier
        :ivar target_port_in_time_range_end: The latest time that the port should occur on the target port in date. Expected format is ISO Offset Time, example: ‘10:15:00-08:00'. We can't guarantee the exact date and time, as this depends on the losing carrier
        :ivar bundle_sid: The bundle sid is an optional identifier to reference a group of regulatory documents for a port request.
        :ivar portability_advance_carrier: A field only required for Japan port in requests. It is a unique identifier for the donor carrier service the line is being ported from.
        :ivar auto_cancel_approval_numbers: Japan specific field, indicates the number of phone numbers to automatically approve for cancellation.
        """

        def __init__(self, payload: Dict[str, Any]):

            self.account_sid: Optional[str] = payload.get("account_sid")
            self.documents: Optional[List[str]] = payload.get("documents")
            self.phone_numbers: Optional[
                List[PortingPortInList.NumbersV1PortingPortInCreatePhoneNumbers]
            ] = payload.get("phone_numbers")
            self.losing_carrier_information: Optional[
                PortingPortInList.NumbersV1PortingLosingCarrierInformation
            ] = payload.get("losing_carrier_information")
            self.notification_emails: Optional[List[str]] = payload.get(
                "notification_emails"
            )
            self.target_port_in_date: Optional[date] = payload.get(
                "target_port_in_date"
            )
            self.target_port_in_time_range_start: Optional[str] = payload.get(
                "target_port_in_time_range_start"
            )
            self.target_port_in_time_range_end: Optional[str] = payload.get(
                "target_port_in_time_range_end"
            )
            self.bundle_sid: Optional[str] = payload.get("bundle_sid")
            self.portability_advance_carrier: Optional[str] = payload.get(
                "portability_advance_carrier"
            )
            self.auto_cancel_approval_numbers: Optional[str] = payload.get(
                "auto_cancel_approval_numbers"
            )

        def to_dict(self):
            return {
                "account_sid": self.account_sid,
                "documents": self.documents,
                "phone_numbers": (
                    [phone_numbers.to_dict() for phone_numbers in self.phone_numbers]
                    if self.phone_numbers is not None
                    else None
                ),
                "losing_carrier_information": (
                    self.losing_carrier_information.to_dict()
                    if self.losing_carrier_information is not None
                    else None
                ),
                "notification_emails": self.notification_emails,
                "target_port_in_date": self.target_port_in_date,
                "target_port_in_time_range_start": self.target_port_in_time_range_start,
                "target_port_in_time_range_end": self.target_port_in_time_range_end,
                "bundle_sid": self.bundle_sid,
                "portability_advance_carrier": self.portability_advance_carrier,
                "auto_cancel_approval_numbers": self.auto_cancel_approval_numbers,
            }

    class NumbersV1PortingPortInCreatePhoneNumbers(object):
        """
        :ivar phone_number: Phone number to be ported. This must be in the E164 Format.
        :ivar pin: Some losing carriers require a PIN to authorize the port of a phone number. If the phone number is a US mobile phone number, the PIN is mandatory to process a porting request. Other carriers and number types may also require a PIN, you'll need to contact the losing carrier to determine what your phone number's PIN is.
        """

        def __init__(self, payload: Dict[str, Any]):

            self.phone_number: Optional[str] = payload.get("phone_number")
            self.pin: Optional[str] = payload.get("pin")

        def to_dict(self):
            return {
                "phone_number": self.phone_number,
                "pin": self.pin,
            }

    def __init__(self, version: Version, port_in_request_sid: str):
        """
        Initialize the PortingPortInContext

        :param version: Version that contains the resource
        :param port_in_request_sid: The SID of the Port In request. This is a unique identifier of the port in request.
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "port_in_request_sid": port_in_request_sid,
        }
        self._uri = "/Porting/PortIn/{port_in_request_sid}".format(**self._solution)

    def delete(self) -> bool:
        """
        Deletes the PortingPortInInstance


        :returns: True if delete succeeds, False otherwise
        """

        headers = values.of({})

        return self._version.delete(method="DELETE", uri=self._uri, headers=headers)

    async def delete_async(self) -> bool:
        """
        Asynchronous coroutine that deletes the PortingPortInInstance


        :returns: True if delete succeeds, False otherwise
        """

        headers = values.of({})

        return await self._version.delete_async(
            method="DELETE", uri=self._uri, headers=headers
        )

    def fetch(self) -> PortingPortInInstance:
        """
        Fetch the PortingPortInInstance


        :returns: The fetched PortingPortInInstance
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = self._version.fetch(method="GET", uri=self._uri, headers=headers)

        return PortingPortInInstance(
            self._version,
            payload,
            port_in_request_sid=self._solution["port_in_request_sid"],
        )

    async def fetch_async(self) -> PortingPortInInstance:
        """
        Asynchronous coroutine to fetch the PortingPortInInstance


        :returns: The fetched PortingPortInInstance
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = await self._version.fetch_async(
            method="GET", uri=self._uri, headers=headers
        )

        return PortingPortInInstance(
            self._version,
            payload,
            port_in_request_sid=self._solution["port_in_request_sid"],
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Numbers.V1.PortingPortInContext {}>".format(context)


class PortingPortInList(ListResource):

    class NumbersV1PortingAddress(object):
        """
        :ivar street: The street address, ex: 101 Spear St
        :ivar street_2: The building information, ex : 5th floor.
        :ivar city: The city name, ex: San Francisco.
        :ivar state: The state name, ex: CA or California. Note this should match the losing carrier’s information exactly. So if they spell out the entire state’s name instead of abbreviating it, please do so.
        :ivar zip: The zip code, ex: 94105.
        :ivar country: The country, ex: USA.
        """

        def __init__(self, payload: Dict[str, Any]):

            self.street: Optional[str] = payload.get("street")
            self.street_2: Optional[str] = payload.get("street_2")
            self.city: Optional[str] = payload.get("city")
            self.state: Optional[str] = payload.get("state")
            self.zip: Optional[str] = payload.get("zip")
            self.country: Optional[str] = payload.get("country")

        def to_dict(self):
            return {
                "street": self.street,
                "street_2": self.street_2,
                "city": self.city,
                "state": self.state,
                "zip": self.zip,
                "country": self.country,
            }

    class NumbersV1PortingLosingCarrierInformation(object):
        """
        :ivar customer_name: Customer name as it is registered with the losing carrier. This can be an individual or a business name depending on the customer type selected.
        :ivar account_number: The account number of the customer for the losing carrier. Only require for mobile phone numbers.
        :ivar account_telephone_number: The account phone number of the customer for the losing carrier.
        :ivar address_sid: If you already have an Address SID that represents the address needed for the LOA, you can provide an Address SID instead of providing the address object in the request body. This will copy the address into the port in request. If changes are made to the Address SID after port in request creation, those changes will not be reflected in the port in request.
        :ivar address:
        :ivar authorized_representative: The first and last name of the person listed with the losing carrier who is authorized to make changes on the account.
        :ivar authorized_representative_email: Email address of the person (owner of the number) who will sign the letter of authorization for the port in request. This email address should belong to the person named in as the authorized representative.
        :ivar customer_type: The type of customer account in the losing carrier. This should either be: 'Individual' or 'Business'.
        :ivar authorized_representative_katakana:
        :ivar sub_municipality:
        :ivar building:
        :ivar katakana_name:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.customer_name: Optional[str] = payload.get("customer_name")
            self.account_number: Optional[str] = payload.get("account_number")
            self.account_telephone_number: Optional[str] = payload.get(
                "account_telephone_number"
            )
            self.address_sid: Optional[str] = payload.get("address_sid")
            self.address: Optional[PortingPortInList.NumbersV1PortingAddress] = (
                payload.get("address")
            )
            self.authorized_representative: Optional[str] = payload.get(
                "authorized_representative"
            )
            self.authorized_representative_email: Optional[str] = payload.get(
                "authorized_representative_email"
            )
            self.customer_type: Optional["PortingPortInInstance.str"] = payload.get(
                "customer_type"
            )
            self.authorized_representative_katakana: Optional[str] = payload.get(
                "authorized_representative_katakana"
            )
            self.sub_municipality: Optional[str] = payload.get("sub_municipality")
            self.building: Optional[str] = payload.get("building")
            self.katakana_name: Optional[str] = payload.get("katakana_name")

        def to_dict(self):
            return {
                "customer_name": self.customer_name,
                "account_number": self.account_number,
                "account_telephone_number": self.account_telephone_number,
                "address_sid": self.address_sid,
                "address": self.address.to_dict() if self.address is not None else None,
                "authorized_representative": self.authorized_representative,
                "authorized_representative_email": self.authorized_representative_email,
                "customer_type": self.customer_type,
                "authorized_representative_katakana": self.authorized_representative_katakana,
                "sub_municipality": self.sub_municipality,
                "building": self.building,
                "katakana_name": self.katakana_name,
            }

    class NumbersV1PortingPortInCreate(object):
        """
        :ivar account_sid: Account Sid or subaccount where the phone number(s) will be Ported
        :ivar documents: List of document SIDs for all phone numbers included in the port in request. At least one document SID referring to a document of the type Utility Bill is required.
        :ivar phone_numbers: List of phone numbers to be ported. Maximum of 1,000 phone numbers per request.
        :ivar losing_carrier_information:
        :ivar notification_emails: Additional emails to send a copy of the signed LOA to.
        :ivar target_port_in_date: Target date to port the number. We cannot guarantee that this date will be honored by the other carriers, please work with Ops to get a confirmation of the firm order commitment (FOC) date. Expected format is ISO Local Date, example: ‘2011-12-03`. This date must be at least 7 days in the future for US ports and 10 days in the future for Japanese ports. We can't guarantee the exact date and time, as this depends on the losing carrier
        :ivar target_port_in_time_range_start: The earliest time that the port should occur on the target port in date. Expected format is ISO Offset Time, example: ‘10:15:00-08:00'. We can't guarantee the exact date and time, as this depends on the losing carrier
        :ivar target_port_in_time_range_end: The latest time that the port should occur on the target port in date. Expected format is ISO Offset Time, example: ‘10:15:00-08:00'. We can't guarantee the exact date and time, as this depends on the losing carrier
        :ivar bundle_sid: The bundle sid is an optional identifier to reference a group of regulatory documents for a port request.
        :ivar portability_advance_carrier: A field only required for Japan port in requests. It is a unique identifier for the donor carrier service the line is being ported from.
        :ivar auto_cancel_approval_numbers: Japan specific field, indicates the number of phone numbers to automatically approve for cancellation.
        """

        def __init__(self, payload: Dict[str, Any]):

            self.account_sid: Optional[str] = payload.get("account_sid")
            self.documents: Optional[List[str]] = payload.get("documents")
            self.phone_numbers: Optional[
                List[PortingPortInList.NumbersV1PortingPortInCreatePhoneNumbers]
            ] = payload.get("phone_numbers")
            self.losing_carrier_information: Optional[
                PortingPortInList.NumbersV1PortingLosingCarrierInformation
            ] = payload.get("losing_carrier_information")
            self.notification_emails: Optional[List[str]] = payload.get(
                "notification_emails"
            )
            self.target_port_in_date: Optional[date] = payload.get(
                "target_port_in_date"
            )
            self.target_port_in_time_range_start: Optional[str] = payload.get(
                "target_port_in_time_range_start"
            )
            self.target_port_in_time_range_end: Optional[str] = payload.get(
                "target_port_in_time_range_end"
            )
            self.bundle_sid: Optional[str] = payload.get("bundle_sid")
            self.portability_advance_carrier: Optional[str] = payload.get(
                "portability_advance_carrier"
            )
            self.auto_cancel_approval_numbers: Optional[str] = payload.get(
                "auto_cancel_approval_numbers"
            )

        def to_dict(self):
            return {
                "account_sid": self.account_sid,
                "documents": self.documents,
                "phone_numbers": (
                    [phone_numbers.to_dict() for phone_numbers in self.phone_numbers]
                    if self.phone_numbers is not None
                    else None
                ),
                "losing_carrier_information": (
                    self.losing_carrier_information.to_dict()
                    if self.losing_carrier_information is not None
                    else None
                ),
                "notification_emails": self.notification_emails,
                "target_port_in_date": self.target_port_in_date,
                "target_port_in_time_range_start": self.target_port_in_time_range_start,
                "target_port_in_time_range_end": self.target_port_in_time_range_end,
                "bundle_sid": self.bundle_sid,
                "portability_advance_carrier": self.portability_advance_carrier,
                "auto_cancel_approval_numbers": self.auto_cancel_approval_numbers,
            }

    class NumbersV1PortingPortInCreatePhoneNumbers(object):
        """
        :ivar phone_number: Phone number to be ported. This must be in the E164 Format.
        :ivar pin: Some losing carriers require a PIN to authorize the port of a phone number. If the phone number is a US mobile phone number, the PIN is mandatory to process a porting request. Other carriers and number types may also require a PIN, you'll need to contact the losing carrier to determine what your phone number's PIN is.
        """

        def __init__(self, payload: Dict[str, Any]):

            self.phone_number: Optional[str] = payload.get("phone_number")
            self.pin: Optional[str] = payload.get("pin")

        def to_dict(self):
            return {
                "phone_number": self.phone_number,
                "pin": self.pin,
            }

    def __init__(self, version: Version):
        """
        Initialize the PortingPortInList

        :param version: Version that contains the resource

        """
        super().__init__(version)

        self._uri = "/Porting/PortIn"

    def create(
        self, numbers_v1_porting_port_in_create: NumbersV1PortingPortInCreate
    ) -> PortingPortInInstance:
        """
        Create the PortingPortInInstance

        :param numbers_v1_porting_port_in_create:

        :returns: The created PortingPortInInstance
        """
        data = numbers_v1_porting_port_in_create.to_dict()

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/json"

        headers["Accept"] = "application/json"

        payload = self._version.create(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return PortingPortInInstance(self._version, payload)

    async def create_async(
        self, numbers_v1_porting_port_in_create: NumbersV1PortingPortInCreate
    ) -> PortingPortInInstance:
        """
        Asynchronously create the PortingPortInInstance

        :param numbers_v1_porting_port_in_create:

        :returns: The created PortingPortInInstance
        """
        data = numbers_v1_porting_port_in_create.to_dict()

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/json"

        headers["Accept"] = "application/json"

        payload = await self._version.create_async(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return PortingPortInInstance(self._version, payload)

    def get(self, port_in_request_sid: str) -> PortingPortInContext:
        """
        Constructs a PortingPortInContext

        :param port_in_request_sid: The SID of the Port In request. This is a unique identifier of the port in request.
        """
        return PortingPortInContext(
            self._version, port_in_request_sid=port_in_request_sid
        )

    def __call__(self, port_in_request_sid: str) -> PortingPortInContext:
        """
        Constructs a PortingPortInContext

        :param port_in_request_sid: The SID of the Port In request. This is a unique identifier of the port in request.
        """
        return PortingPortInContext(
            self._version, port_in_request_sid=port_in_request_sid
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Numbers.V1.PortingPortInList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Preview
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from datetime import datetime
from typing import Any, Dict, List, Optional, Union, Iterator, AsyncIterator
from twilio.base import deserialize, serialize, values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version
from twilio.base.page import Page
from twilio.rest.preview.hosted_numbers.authorization_document.dependent_hosted_number_order import (
    DependentHostedNumberOrderList,
)


class AuthorizationDocumentInstance(InstanceResource):

    class Status(object):
        OPENED = "opened"
        SIGNING = "signing"
        SIGNED = "signed"
        CANCELED = "canceled"
        FAILED = "failed"

    """
    :ivar sid: A 34 character string that uniquely identifies this AuthorizationDocument.
    :ivar address_sid: A 34 character string that uniquely identifies the Address resource that is associated with this AuthorizationDocument.
    :ivar status: 
    :ivar email: Email that this AuthorizationDocument will be sent to for signing.
    :ivar cc_emails: Email recipients who will be informed when an Authorization Document has been sent and signed.
    :ivar date_created: The date this resource was created, given as [GMT RFC 2822](http://www.ietf.org/rfc/rfc2822.txt) format.
    :ivar date_updated: The date that this resource was updated, given as [GMT RFC 2822](http://www.ietf.org/rfc/rfc2822.txt) format.
    :ivar url: 
    :ivar links: 
    """

    def __init__(
        self, version: Version, payload: Dict[str, Any], sid: Optional[str] = None
    ):
        super().__init__(version)

        self.sid: Optional[str] = payload.get("sid")
        self.address_sid: Optional[str] = payload.get("address_sid")
        self.status: Optional["AuthorizationDocumentInstance.Status"] = payload.get(
            "status"
        )
        self.email: Optional[str] = payload.get("email")
        self.cc_emails: Optional[List[str]] = payload.get("cc_emails")
        self.date_created: Optional[datetime] = deserialize.iso8601_datetime(
            payload.get("date_created")
        )
        self.date_updated: Optional[datetime] = deserialize.iso8601_datetime(
            payload.get("date_updated")
        )
        self.url: Optional[str] = payload.get("url")
        self.links: Optional[Dict[str, object]] = payload.get("links")

        self._solution = {
            "sid": sid or self.sid,
        }
        self._context: Optional[AuthorizationDocumentContext] = None

    @property
    def _proxy(self) -> "AuthorizationDocumentContext":
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: AuthorizationDocumentContext for this AuthorizationDocumentInstance
        """
        if self._context is None:
            self._context = AuthorizationDocumentContext(
                self._version,
                sid=self._solution["sid"],
            )
        return self._context

    def fetch(self) -> "AuthorizationDocumentInstance":
        """
        Fetch the AuthorizationDocumentInstance


        :returns: The fetched AuthorizationDocumentInstance
        """
        return self._proxy.fetch()

    async def fetch_async(self) -> "AuthorizationDocumentInstance":
        """
        Asynchronous coroutine to fetch the AuthorizationDocumentInstance


        :returns: The fetched AuthorizationDocumentInstance
        """
        return await self._proxy.fetch_async()

    def update(
        self,
        hosted_number_order_sids: Union[List[str], object] = values.unset,
        address_sid: Union[str, object] = values.unset,
        email: Union[str, object] = values.unset,
        cc_emails: Union[List[str], object] = values.unset,
        status: Union["AuthorizationDocumentInstance.Status", object] = values.unset,
        contact_title: Union[str, object] = values.unset,
        contact_phone_number: Union[str, object] = values.unset,
    ) -> "AuthorizationDocumentInstance":
        """
        Update the AuthorizationDocumentInstance

        :param hosted_number_order_sids: A list of HostedNumberOrder sids that this AuthorizationDocument will authorize for hosting phone number capabilities on Twilio's platform.
        :param address_sid: A 34 character string that uniquely identifies the Address resource that is associated with this AuthorizationDocument.
        :param email: Email that this AuthorizationDocument will be sent to for signing.
        :param cc_emails: Email recipients who will be informed when an Authorization Document has been sent and signed
        :param status:
        :param contact_title: The title of the person authorized to sign the Authorization Document for this phone number.
        :param contact_phone_number: The contact phone number of the person authorized to sign the Authorization Document.

        :returns: The updated AuthorizationDocumentInstance
        """
        return self._proxy.update(
            hosted_number_order_sids=hosted_number_order_sids,
            address_sid=address_sid,
            email=email,
            cc_emails=cc_emails,
            status=status,
            contact_title=contact_title,
            contact_phone_number=contact_phone_number,
        )

    async def update_async(
        self,
        hosted_number_order_sids: Union[List[str], object] = values.unset,
        address_sid: Union[str, object] = values.unset,
        email: Union[str, object] = values.unset,
        cc_emails: Union[List[str], object] = values.unset,
        status: Union["AuthorizationDocumentInstance.Status", object] = values.unset,
        contact_title: Union[str, object] = values.unset,
        contact_phone_number: Union[str, object] = values.unset,
    ) -> "AuthorizationDocumentInstance":
        """
        Asynchronous coroutine to update the AuthorizationDocumentInstance

        :param hosted_number_order_sids: A list of HostedNumberOrder sids that this AuthorizationDocument will authorize for hosting phone number capabilities on Twilio's platform.
        :param address_sid: A 34 character string that uniquely identifies the Address resource that is associated with this AuthorizationDocument.
        :param email: Email that this AuthorizationDocument will be sent to for signing.
        :param cc_emails: Email recipients who will be informed when an Authorization Document has been sent and signed
        :param status:
        :param contact_title: The title of the person authorized to sign the Authorization Document for this phone number.
        :param contact_phone_number: The contact phone number of the person authorized to sign the Authorization Document.

        :returns: The updated AuthorizationDocumentInstance
        """
        return await self._proxy.update_async(
            hosted_number_order_sids=hosted_number_order_sids,
            address_sid=address_sid,
            email=email,
            cc_emails=cc_emails,
            status=status,
            contact_title=contact_title,
            contact_phone_number=contact_phone_number,
        )

    @property
    def dependent_hosted_number_orders(self) -> DependentHostedNumberOrderList:
        """
        Access the dependent_hosted_number_orders
        """
        return self._proxy.dependent_hosted_number_orders

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Preview.HostedNumbers.AuthorizationDocumentInstance {}>".format(
            context
        )


class AuthorizationDocumentContext(InstanceContext):

    def __init__(self, version: Version, sid: str):
        """
        Initialize the AuthorizationDocumentContext

        :param version: Version that contains the resource
        :param sid: A 34 character string that uniquely identifies this AuthorizationDocument.
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "sid": sid,
        }
        self._uri = "/AuthorizationDocuments/{sid}".format(**self._solution)

        self._dependent_hosted_number_orders: Optional[
            DependentHostedNumberOrderList
        ] = None

    def fetch(self) -> AuthorizationDocumentInstance:
        """
        Fetch the AuthorizationDocumentInstance


        :returns: The fetched AuthorizationDocumentInstance
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = self._version.fetch(method="GET", uri=self._uri, headers=headers)

        return AuthorizationDocumentInstance(
            self._version,
            payload,
            sid=self._solution["sid"],
        )

    async def fetch_async(self) -> AuthorizationDocumentInstance:
        """
        Asynchronous coroutine to fetch the AuthorizationDocumentInstance


        :returns: The fetched AuthorizationDocumentInstance
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = await self._version.fetch_async(
            method="GET", uri=self._uri, headers=headers
        )

        return AuthorizationDocumentInstance(
            self._version,
            payload,
            sid=self._solution["sid"],
        )

    def update(
        self,
        hosted_number_order_sids: Union[List[str], object] = values.unset,
        address_sid: Union[str, object] = values.unset,
        email: Union[str, object] = values.unset,
        cc_emails: Union[List[str], object] = values.unset,
        status: Union["AuthorizationDocumentInstance.Status", object] = values.unset,
        contact_title: Union[str, object] = values.unset,
        contact_phone_number: Union[str, object] = values.unset,
    ) -> AuthorizationDocumentInstance:
        """
        Update the AuthorizationDocumentInstance

        :param hosted_number_order_sids: A list of HostedNumberOrder sids that this AuthorizationDocument will authorize for hosting phone number capabilities on Twilio's platform.
        :param address_sid: A 34 character string that uniquely identifies the Address resource that is associated with this AuthorizationDocument.
        :param email: Email that this AuthorizationDocument will be sent to for signing.
        :param cc_emails: Email recipients who will be informed when an Authorization Document has been sent and signed
        :param status:
        :param contact_title: The title of the person authorized to sign the Authorization Document for this phone number.
        :param contact_phone_number: The contact phone number of the person authorized to sign the Authorization Document.

        :returns: The updated AuthorizationDocumentInstance
        """

        data = values.of(
            {
                "HostedNumberOrderSids": serialize.map(
                    hosted_number_order_sids, lambda e: e
                ),
                "AddressSid": address_sid,
                "Email": email,
                "CcEmails": serialize.map(cc_emails, lambda e: e),
                "Status": status,
                "ContactTitle": contact_title,
                "ContactPhoneNumber": contact_phone_number,
            }
        )
        headers = values.of({})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = self._version.update(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return AuthorizationDocumentInstance(
            self._version, payload, sid=self._solution["sid"]
        )

    async def update_async(
        self,
        hosted_number_order_sids: Union[List[str], object] = values.unset,
        address_sid: Union[str, object] = values.unset,
        email: Union[str, object] = values.unset,
        cc_emails: Union[List[str], object] = values.unset,
        status: Union["AuthorizationDocumentInstance.Status", object] = values.unset,
        contact_title: Union[str, object] = values.unset,
        contact_phone_number: Union[str, object] = values.unset,
    ) -> AuthorizationDocumentInstance:
        """
        Asynchronous coroutine to update the AuthorizationDocumentInstance

        :param hosted_number_order_sids: A list of HostedNumberOrder sids that this AuthorizationDocument will authorize for hosting phone number capabilities on Twilio's platform.
        :param address_sid: A 34 character string that uniquely identifies the Address resource that is associated with this AuthorizationDocument.
        :param email: Email that this AuthorizationDocument will be sent to for signing.
        :param cc_emails: Email recipients who will be informed when an Authorization Document has been sent and signed
        :param status:
        :param contact_title: The title of the person authorized to sign the Authorization Document for this phone number.
        :param contact_phone_number: The contact phone number of the person authorized to sign the Authorization Document.

        :returns: The updated AuthorizationDocumentInstance
        """

        data = values.of(
            {
                "HostedNumberOrderSids": serialize.map(
                    hosted_number_order_sids, lambda e: e
                ),
                "AddressSid": address_sid,
                "Email": email,
                "CcEmails": serialize.map(cc_emails, lambda e: e),
                "Status": status,
                "ContactTitle": contact_title,
                "ContactPhoneNumber": contact_phone_number,
            }
        )
        headers = values.of({})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = await self._version.update_async(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return AuthorizationDocumentInstance(
            self._version, payload, sid=self._solution["sid"]
        )

    @property
    def dependent_hosted_number_orders(self) -> DependentHostedNumberOrderList:
        """
        Access the dependent_hosted_number_orders
        """
        if self._dependent_hosted_number_orders is None:
            self._dependent_hosted_number_orders = DependentHostedNumberOrderList(
                self._version,
                self._solution["sid"],
            )
        return self._dependent_hosted_number_orders

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Preview.HostedNumbers.AuthorizationDocumentContext {}>".format(
            context
        )


class AuthorizationDocumentPage(Page):

    def get_instance(self, payload: Dict[str, Any]) -> AuthorizationDocumentInstance:
        """
        Build an instance of AuthorizationDocumentInstance

        :param payload: Payload response from the API
        """
        return AuthorizationDocumentInstance(self._version, payload)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Preview.HostedNumbers.AuthorizationDocumentPage>"


class AuthorizationDocumentList(ListResource):

    def __init__(self, version: Version):
        """
        Initialize the AuthorizationDocumentList

        :param version: Version that contains the resource

        """
        super().__init__(version)

        self._uri = "/AuthorizationDocuments"

    def create(
        self,
        hosted_number_order_sids: List[str],
        address_sid: str,
        email: str,
        contact_title: str,
        contact_phone_number: str,
        cc_emails: Union[List[str], object] = values.unset,
    ) -> AuthorizationDocumentInstance:
        """
        Create the AuthorizationDocumentInstance

        :param hosted_number_order_sids: A list of HostedNumberOrder sids that this AuthorizationDocument will authorize for hosting phone number capabilities on Twilio's platform.
        :param address_sid: A 34 character string that uniquely identifies the Address resource that is associated with this AuthorizationDocument.
        :param email: Email that this AuthorizationDocument will be sent to for signing.
        :param contact_title: The title of the person authorized to sign the Authorization Document for this phone number.
        :param contact_phone_number: The contact phone number of the person authorized to sign the Authorization Document.
        :param cc_emails: Email recipients who will be informed when an Authorization Document has been sent and signed.

        :returns: The created AuthorizationDocumentInstance
        """

        data = values.of(
            {
                "HostedNumberOrderSids": serialize.map(
                    hosted_number_order_sids, lambda e: e
                ),
                "AddressSid": address_sid,
                "Email": email,
                "ContactTitle": contact_title,
                "ContactPhoneNumber": contact_phone_number,
                "CcEmails": serialize.map(cc_emails, lambda e: e),
            }
        )
        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = self._version.create(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return AuthorizationDocumentInstance(self._version, payload)

    async def create_async(
        self,
        hosted_number_order_sids: List[str],
        address_sid: str,
        email: str,
        contact_title: str,
        contact_phone_number: str,
        cc_emails: Union[List[str], object] = values.unset,
    ) -> AuthorizationDocumentInstance:
        """
        Asynchronously create the AuthorizationDocumentInstance

        :param hosted_number_order_sids: A list of HostedNumberOrder sids that this AuthorizationDocument will authorize for hosting phone number capabilities on Twilio's platform.
        :param address_sid: A 34 character string that uniquely identifies the Address resource that is associated with this AuthorizationDocument.
        :param email: Email that this AuthorizationDocument will be sent to for signing.
        :param contact_title: The title of the person authorized to sign the Authorization Document for this phone number.
        :param contact_phone_number: The contact phone number of the person authorized to sign the Authorization Document.
        :param cc_emails: Email recipients who will be informed when an Authorization Document has been sent and signed.

        :returns: The created AuthorizationDocumentInstance
        """

        data = values.of(
            {
                "HostedNumberOrderSids": serialize.map(
                    hosted_number_order_sids, lambda e: e
                ),
                "AddressSid": address_sid,
                "Email": email,
                "ContactTitle": contact_title,
                "ContactPhoneNumber": contact_phone_number,
                "CcEmails": serialize.map(cc_emails, lambda e: e),
            }
        )
        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = await self._version.create_async(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return AuthorizationDocumentInstance(self._version, payload)

    def stream(
        self,
        email: Union[str, object] = values.unset,
        status: Union["AuthorizationDocumentInstance.Status", object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Iterator[AuthorizationDocumentInstance]:
        """
        Streams AuthorizationDocumentInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param str email: Email that this AuthorizationDocument will be sent to for signing.
        :param &quot;AuthorizationDocumentInstance.Status&quot; status: Status of an instance resource. It can hold one of the values: 1. opened 2. signing, 3. signed LOA, 4. canceled, 5. failed. See the section entitled [Status Values](https://www.twilio.com/docs/phone-numbers/hosted-numbers/hosted-numbers-api/authorization-document-resource#status-values) for more information on each of these statuses.
        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = self.page(email=email, status=status, page_size=limits["page_size"])

        return self._version.stream(page, limits["limit"])

    async def stream_async(
        self,
        email: Union[str, object] = values.unset,
        status: Union["AuthorizationDocumentInstance.Status", object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> AsyncIterator[AuthorizationDocumentInstance]:
        """
        Asynchronously streams AuthorizationDocumentInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param str email: Email that this AuthorizationDocument will be sent to for signing.
        :param &quot;AuthorizationDocumentInstance.Status&quot; status: Status of an instance resource. It can hold one of the values: 1. opened 2. signing, 3. signed LOA, 4. canceled, 5. failed. See the section entitled [Status Values](https://www.twilio.com/docs/phone-numbers/hosted-numbers/hosted-numbers-api/authorization-document-resource#status-values) for more information on each of these statuses.
        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = await self.page_async(
            email=email, status=status, page_size=limits["page_size"]
        )

        return self._version.stream_async(page, limits["limit"])

    def list(
        self,
        email: Union[str, object] = values.unset,
        status: Union["AuthorizationDocumentInstance.Status", object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[AuthorizationDocumentInstance]:
        """
        Lists AuthorizationDocumentInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param str email: Email that this AuthorizationDocument will be sent to for signing.
        :param &quot;AuthorizationDocumentInstance.Status&quot; status: Status of an instance resource. It can hold one of the values: 1. opened 2. signing, 3. signed LOA, 4. canceled, 5. failed. See the section entitled [Status Values](https://www.twilio.com/docs/phone-numbers/hosted-numbers/hosted-numbers-api/authorization-document-resource#status-values) for more information on each of these statuses.
        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return list(
            self.stream(
                email=email,
                status=status,
                limit=limit,
                page_size=page_size,
            )
        )

    async def list_async(
        self,
        email: Union[str, object] = values.unset,
        status: Union["AuthorizationDocumentInstance.Status", object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[AuthorizationDocumentInstance]:
        """
        Asynchronously lists AuthorizationDocumentInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param str email: Email that this AuthorizationDocument will be sent to for signing.
        :param &quot;AuthorizationDocumentInstance.Status&quot; status: Status of an instance resource. It can hold one of the values: 1. opened 2. signing, 3. signed LOA, 4. canceled, 5. failed. See the section entitled [Status Values](https://www.twilio.com/docs/phone-numbers/hosted-numbers/hosted-numbers-api/authorization-document-resource#status-values) for more information on each of these statuses.
        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return [
            record
            async for record in await self.stream_async(
                email=email,
                status=status,
                limit=limit,
                page_size=page_size,
            )
        ]

    def page(
        self,
        email: Union[str, object] = values.unset,
        status: Union["AuthorizationDocumentInstance.Status", object] = values.unset,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> AuthorizationDocumentPage:
        """
        Retrieve a single page of AuthorizationDocumentInstance records from the API.
        Request is executed immediately

        :param email: Email that this AuthorizationDocument will be sent to for signing.
        :param status: Status of an instance resource. It can hold one of the values: 1. opened 2. signing, 3. signed LOA, 4. canceled, 5. failed. See the section entitled [Status Values](https://www.twilio.com/docs/phone-numbers/hosted-numbers/hosted-numbers-api/authorization-document-resource#status-values) for more information on each of these statuses.
        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of AuthorizationDocumentInstance
        """
        data = values.of(
            {
                "Email": email,
                "Status": status,
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response = self._version.page(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return AuthorizationDocumentPage(self._version, response)

    async def page_async(
        self,
        email: Union[str, object] = values.unset,
        status: Union["AuthorizationDocumentInstance.Status", object] = values.unset,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> AuthorizationDocumentPage:
        """
        Asynchronously retrieve a single page of AuthorizationDocumentInstance records from the API.
        Request is executed immediately

        :param email: Email that this AuthorizationDocument will be sent to for signing.
        :param status: Status of an instance resource. It can hold one of the values: 1. opened 2. signing, 3. signed LOA, 4. canceled, 5. failed. See the section entitled [Status Values](https://www.twilio.com/docs/phone-numbers/hosted-numbers/hosted-numbers-api/authorization-document-resource#status-values) for more information on each of these statuses.
        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of AuthorizationDocumentInstance
        """
        data = values.of(
            {
                "Email": email,
                "Status": status,
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response = await self._version.page_async(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return AuthorizationDocumentPage(self._version, response)

    def get_page(self, target_url: str) -> AuthorizationDocumentPage:
        """
        Retrieve a specific page of AuthorizationDocumentInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of AuthorizationDocumentInstance
        """
        response = self._version.domain.twilio.request("GET", target_url)
        return AuthorizationDocumentPage(self._version, response)

    async def get_page_async(self, target_url: str) -> AuthorizationDocumentPage:
        """
        Asynchronously retrieve a specific page of AuthorizationDocumentInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of AuthorizationDocumentInstance
        """
        response = await self._version.domain.twilio.request_async("GET", target_url)
        return AuthorizationDocumentPage(self._version, response)

    def get(self, sid: str) -> AuthorizationDocumentContext:
        """
        Constructs a AuthorizationDocumentContext

        :param sid: A 34 character string that uniquely identifies this AuthorizationDocument.
        """
        return AuthorizationDocumentContext(self._version, sid=sid)

    def __call__(self, sid: str) -> AuthorizationDocumentContext:
        """
        Constructs a AuthorizationDocumentContext

        :param sid: A 34 character string that uniquely identifies this AuthorizationDocument.
        """
        return AuthorizationDocumentContext(self._version, sid=sid)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Preview.HostedNumbers.AuthorizationDocumentList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Studio
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from typing import Optional
from twilio.base.version import Version
from twilio.base.domain import Domain
from twilio.rest.studio.v2.flow import FlowList
from twilio.rest.studio.v2.flow_validate import FlowValidateList


class V2(Version):

    def __init__(self, domain: Domain):
        """
        Initialize the V2 version of Studio

        :param domain: The Twilio.studio domain
        """
        super().__init__(domain, "v2")
        self._flows: Optional[FlowList] = None
        self._flow_validate: Optional[FlowValidateList] = None

    @property
    def flows(self) -> FlowList:
        if self._flows is None:
            self._flows = FlowList(self)
        return self._flows

    @property
    def flow_validate(self) -> FlowValidateList:
        if self._flow_validate is None:
            self._flow_validate = FlowValidateList(self)
        return self._flow_validate

    def __repr__(self) -> str:
        """
        Provide a friendly representation
        :returns: Machine friendly representation
        """
        return "<Twilio.Studio.V2>"
